from typing import Any
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def update_vpc_endpoint(
    hub,
    ctx,
    old_state: Dict[str, Any],
    resource_id: str,
    policy_document: str = None,
    route_table_ids: List = None,
    subnet_ids: List = None,
    security_group_ids: List = None,
    private_dns_enabled: bool = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    """
    Modifies a vpc endpoint. The values subnet_ids, security_group_ids, route_table_ids will actually become
    the respective values for the vpc endpoint. It will error if you try to enable private dns on a non Interface endpoint.

    Args:
        hub:
        ctx:
        old_state: Dict[Text, Any]: The previous state of the vpc_endpoint.
        resource_id(string): AWS VPC Endpoint id.
        policy_document(string): The new policy document for the vpc endpoint
        route_table_ids(List): List of route table ids for the vpc endpoint
        subnet_ids(List): List of subnet ids for the vpc endpoint
        security_group_ids(List): List of security group ids for the vpc endpoint
        private_dns_enabled(bool): Whether to enable or disable private dns for the vpc endpoint
        timeout(Dict, optional): Timeout configuration for updating the endpoint.
            * update (Dict) -- Timeout configuration for updating the endpoint
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

    Returns:
        {"result": True|False, "comment": "A message tuple", "ret": Dict}

    """

    result = dict(comment=(), result=True, ret=None)
    name = old_state.get("name")

    # For if the vpc_endpoint exists
    hub.log.debug(f"Updating aws.ec2.vpc_endpoint: '{name}'")

    if old_state.get("policy_document") and policy_document == None:
        policy_document = old_state["policy_document"]

    if private_dns_enabled == None:
        private_dns_enabled = old_state["private_dns_enabled"]

    is_update_required = (
        policy_document and policy_document != old_state["policy_document"]
    ) or (private_dns_enabled != old_state["private_dns_enabled"])

    find_set_difference = lambda x, y, z: (
        list(set(x).difference(set(y))),
        (z or len(set(x).difference(set(y))) > 0),
    )

    if route_table_ids is None:
        route_tables_to_be_removed = []
        new_route_tables = []
    elif old_state.get("route_table_ids"):
        # find the route tables in route_table_ids that aren't in the old state (these are the new ones)
        new_route_tables, is_update_required = find_set_difference(
            route_table_ids, old_state["route_table_ids"], is_update_required
        )
        # the ones in the old_state not in route_table_ids are the ones to be removed
        route_tables_to_be_removed, is_update_required = find_set_difference(
            old_state["route_table_ids"], route_table_ids, is_update_required
        )
    else:
        new_route_tables = route_table_ids
        route_tables_to_be_removed = old_state.get("route_table_ids", [])
        is_update_required = is_update_required or len(route_table_ids) > 0
    hub.log.debug(
        f"Updating aws.ec2.vpc_endpoint: '{name}'. Route tables passed in: {route_table_ids}. Old route tables: {old_state.get('route_table_ids', None)}"
    )
    hub.log.debug(
        f"Updating aws.ec2.vpc_endpoint: '{name}'. New route tables: {new_route_tables}. Route tables to be removed: {route_tables_to_be_removed}"
    )

    # The logic for subnets and security group's is the same as Route tables
    if subnet_ids is None:
        subnets_to_be_removed = []
        new_subnets = []
    elif old_state.get("subnet_ids"):
        new_subnets, is_update_required = find_set_difference(
            subnet_ids, old_state["subnet_ids"], is_update_required
        )
        subnets_to_be_removed, is_update_required = find_set_difference(
            old_state["subnet_ids"], subnet_ids, is_update_required
        )
    else:
        new_subnets = subnet_ids
        subnets_to_be_removed = old_state.get("subnet_ids", [])
        is_update_required = is_update_required or len(subnet_ids) > 0
    hub.log.debug(
        f"Updating aws.ec2.vpc_endpoint: '{name}'. Subnets passed in: {subnet_ids}. Old subnets: {old_state.get('subnet_ids', None)}"
    )
    hub.log.debug(
        f"Updating aws.ec2.vpc_endpoint: '{name}'. New subnets: {new_subnets}. Subnets to be removed: {subnets_to_be_removed}"
    )

    if security_group_ids is None:
        security_groups_to_be_removed = []
        new_security_groups = []
    elif old_state.get("security_group_ids"):
        new_security_groups, is_update_required = find_set_difference(
            security_group_ids, old_state["security_group_ids"], is_update_required
        )
        security_groups_to_be_removed, is_update_required = find_set_difference(
            old_state["security_group_ids"], security_group_ids, is_update_required
        )
    else:
        new_security_groups = security_group_ids
        security_groups_to_be_removed = old_state.get("security_group_ids", [])
        is_update_required = is_update_required or len(security_group_ids) > 0
    hub.log.debug(
        f"Updating aws.ec2.vpc_endpoint: '{name}'. Security groups passed in: {security_group_ids}. Old security groups: {(old_state.get('security_group_ids', None))}"
    )
    hub.log.debug(
        f"Updating aws.ec2.vpc_endpoint: '{name}'. New security groups: {new_security_groups}. Security groups to be removed: {security_groups_to_be_removed}"
    )

    if not is_update_required:
        return result

    if ctx.get("test", False) and result["result"]:
        plan_state = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state=old_state,
            desired_state={
                "policy_document": policy_document,
                "route_table_ids": route_table_ids,
                "subnet_ids": subnet_ids,
                "security_group_ids": security_group_ids,
                "private_dns_enabled": private_dns_enabled,
            },
        )
        result["ret"] = {k: v for k, v in plan_state.items() if v is not None}
        result["comment"] = (
            f"Would succeed in modifying aws.ec2.vpc_endpoint '{name}'",
        )
        return result

    ret = await hub.exec.boto3.client.ec2.modify_vpc_endpoint(
        ctx,
        VpcEndpointId=resource_id,
        ResetPolicy=False,
        PolicyDocument=policy_document,
        AddRouteTableIds=new_route_tables,
        RemoveRouteTableIds=route_tables_to_be_removed,
        AddSubnetIds=new_subnets,
        RemoveSubnetIds=subnets_to_be_removed,
        AddSecurityGroupIds=new_security_groups,
        RemoveSecurityGroupIds=security_groups_to_be_removed,
        PrivateDnsEnabled=private_dns_enabled,
    )
    result["result"] = ret["result"]
    result["ret"] = ret["ret"]
    if not result["result"]:
        result["comment"] = result["comment"] + ret["comment"]
        return result
    update_waiter_acceptors = [
        {
            "matcher": "pathAll",
            "expected": "available",
            "state": "success",
            "argument": "VpcEndpoints[].State",
        },
        {
            "matcher": "pathAll",
            "expected": "pending",
            "state": "retry",
            "argument": "VpcEndpoints[].State",
        },
    ]
    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=15,
        default_max_attempts=40,
        timeout_config=timeout.get("update") if timeout else None,
    )
    endpoint_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
        name="VpcEndpointModify",
        operation="DescribeVpcEndpoints",
        argument=["VpcEndpoints[].State"],
        acceptors=update_waiter_acceptors,
        client=hub.tool.boto3.client.get_client(ctx, "ec2"),
    )
    try:
        await hub.tool.boto3.client.wait(
            ctx,
            "ec2",
            "VpcEndpointModify",
            endpoint_waiter,
            VpcEndpointIds=[resource_id],
            WaiterConfig=waiter_config,
        )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
        return result

    result["comment"] = (f"Succeeded modifying aws.ec2.vpc_endpoint '{name}'",)
    return result


async def get(
    hub,
    ctx,
    name,
    resource_id: str,
) -> Dict:
    """
    Use an un-managed vpc_endpoint as a data-source. Supply resource_id as a filter.

    Args:
        name(Text): The name of the Idem state.
        resource_id(Text):  An AWS vpc_endpoint resource_id to identify the resource.
    """

    result = dict(comment=[], ret=None, result=True)

    resource_ret = await hub.exec.boto3.client.ec2.describe_vpc_endpoints(
        ctx, VpcEndpointIds=[resource_id]
    )

    if not resource_ret["result"]:
        if "InvalidVpcEndpointId.NotFound" in str(resource_ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.vpc_endpoint", name=name
                )
            )
            result["comment"] += list(resource_ret["comment"])
            return result
        result["comment"] += list(resource_ret["comment"])
        result["result"] = resource_ret["result"]
        return result

    if not resource_ret["ret"]["VpcEndpoints"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.vpc_endpoint", name=name
            )
        )
        return result

    result[
        "ret"
    ] = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_endpoint_to_present(
        raw_resource=resource_ret["ret"]["VpcEndpoints"][0], idem_resource_name=name
    )

    return result


async def list_(hub, ctx, name) -> Dict:
    """
    Use an un-managed VPC Endpoints as a data-source.

    Args:
        name(string): The name of the Idem state.

    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.exec.boto3.client.ec2.describe_vpc_endpoints(ctx)
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["VpcEndpoints"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.vpc_endpoint", name=name
            )
        )
        return result
    for vpc_endpoint in ret["ret"]["VpcEndpoints"]:
        resource_id = vpc_endpoint.get("VpcEndpointId")
        converted_resource = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_endpoint_to_present(
                vpc_endpoint, idem_resource_name=resource_id
            )
        )
        result["ret"].append(converted_resource)
    return result



import copy
from typing import Dict


async def update(
    hub,
    ctx,
    resource_id,
    old_tags: Dict[str, str],
    new_tags: Dict[str, str],
):
    """
    Update tags of AWS ElasticLoadBalancing resources.

    Args:
        resource_id(Text): AWS ElasticLoadBalancing resource id
        old_tags(Dict[str, str]): dict of old tags in the format of {"Name": name, "Description": Description}
        new_tags(Dict[str, str]): dict of new tags in the format of {"Name": name, "Description": Description}.
            If this value is None, the function will do no operation on tags.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": dict of updated tags}

    """
    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret={})
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.elb.remove_tags(
                ctx,
                LoadBalancerNames=[resource_id],
                Tags=[{"Key": key} for key, value in tags_to_remove.items()],
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.elb.add_tags(
                ctx,
                LoadBalancerNames=[resource_id],
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags=tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] += add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result



from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    filters: List = None,
    tags: Dict[str, str] = None,
) -> Dict:
    """Get a single Elastic IP address from AWS.

    If more than one resource is found, the first resource returned from AWS will be used. The function returns None
    when no resource is found.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str, Optional):
            Public IP of the Elastic IP address.

        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_addresses

        tags(dict, Optional):
            The dict of tags to filter by. For example, to find all resources that have a tag with the key "Owner" and
            the value "TeamA" , specify "tag:Owner" for the Dict key and "TeamA" for the Dict value.
    """
    result = dict(comment=[], ret=None, result=True)

    tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)

    ret = await hub.tool.aws.ec2.elastic_ip.search_raw(
        ctx=ctx, resource_id=resource_id, filters=filters, tags=tags
    )
    if not ret["result"]:
        if "InvalidAddress.NotFound" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.elastic_ip", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Addresses"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.elastic_ip", name=name
            )
        )
        return result

    resource = ret["ret"]["Addresses"][0]
    if len(ret["ret"]["Addresses"]) > 1:
        result["comment"].append(
            f"More than one aws.ec2.elastic_ip resource was found. Use resource {resource.get('PublicIp')}"
        )
    result["ret"] = hub.tool.aws.ec2.conversion_utils.convert_raw_elastic_ip_to_present(
        raw_resource=resource, idem_resource_name=name
    )

    return result


async def list_(
    hub,
    ctx,
    name: str,
    filters: List = None,
    tags: Dict[str, str] = None,
) -> Dict:
    """Fetch a list of Elastic IP addresses from AWS.

    The function returns empty list when no resource is found.

    Args:
        name(str):
            The name of the Idem state.

        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_addresses

        tags(dict, Optional):
            The dict of tags to filter by. For example, to find all resources that have a tag with the key "Owner" and
            the value "TeamA" , specify "tag:Owner" for the Dict key and "TeamA" for the Dict value.
    """
    result = dict(comment=[], ret=[], result=True)

    tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)

    ret = await hub.tool.aws.ec2.elastic_ip.search_raw(
        ctx=ctx, filters=filters, tags=tags
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Addresses"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.elastic_ip", name=name
            )
        )
        return result

    for address in ret["ret"]["Addresses"]:
        result["ret"].append(
            hub.tool.aws.ec2.conversion_utils.convert_raw_elastic_ip_to_present(
                raw_resource=address, idem_resource_name=address.get("PublicIp")
            )
        )

    return result



import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    rest_api_id: str,
    http_method: str,
    status_code: str,
    parent_resource_id: str,
    resource_id: str = None,
    response_models: Dict = None,
    response_parameters: Dict = None,
):
    """
    Creates a new API Gateway Method Response or modifies an existing one.

    Args:
        name(string):
            An idem name of the resource.

        rest_api_id(string):
            AWS rest_api id of the associated RestApi.

        parent_resource_id(string):
            The parent resource's id.

        http_method(string):
            String that specifies the method request's HTTP method type.

        status_code(string):
            The method response's status code.

        resource_id(string, optional):
            Defaults to None. Idem Resource id, formatted as: rest_api_id-parent_resource_id-http_method-status_code.

        response_models(dict, optional):
            Specifies the Model resources used for the response's content-type.
                Response models are represented as a key/value map, with a content-type as the key and a Model name
                as the value.

        response_parameters(dict, optional):
            A key-value map specifying required or optional response parameters
                that API Gateway can send back to the caller. A key defines a method response header and the value
                specifies whether the associated method response header is required or not. The expression of the key
                must match the pattern method.response.header.{name} , where name is a valid and unique header name.
                API Gateway passes certain integration response data to the method response headers specified here
                according to the mapping you prescribe in the API's IntegrationResponse. The integration response data
                that can be mapped include an integration response header expressed in integration.response.header.{name},
                a static value enclosed within a pair of single quotes (e.g., 'application/json' ), or a JSON expression
                from the back-end response payload in the form of integration.response.body.{JSON-expression},
                where JSON-expression is a valid JSON expression without the $ prefix.)

    Request Syntax:
        [idem_test_aws_apigateway_method_response]:
          aws.apigateway.method_response.present:
            - name: 'string'
            - rest_api_id: 'string'
            - parent_resource_id: 'string'
            - http_method: 'string'
            - status_code: 'string'

    Returns:
        Dict[str, Any]

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.apigateway.method_response.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )

        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

        result["old_state"] = before["ret"]
        result["new_state"] = copy.deepcopy(result["old_state"])

        update_parameters = {
            "response_models": response_models,
            "response_parameters": response_parameters,
        }

        update_method_response_ret = (
            await hub.tool.aws.apigateway.method_response.update(
                ctx,
                old_state=result["old_state"],
                updateable_resource_parameters=update_parameters,
            )
        )
        result["comment"] = result["comment"] + update_method_response_ret["comment"]

        if not update_method_response_ret["result"]:
            result["result"] = False
            return result
        resource_updated = bool(update_method_response_ret["ret"])

    else:
        resource_id = f"{rest_api_id}-{parent_resource_id}-{http_method}-{status_code}"

        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "rest_api_id": rest_api_id,
                    "parent_resource_id": parent_resource_id,
                    "http_method": http_method,
                    "status_code": status_code,
                    "response_parameters": response_parameters,
                    "response_models": response_models,
                },
            )

            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigateway.method_response", name=name
            )
            return result

        ret = await hub.exec.boto3.client.apigateway.put_method_response(
            ctx,
            restApiId=rest_api_id,
            resourceId=parent_resource_id,
            httpMethod=http_method,
            statusCode=status_code,
            responseModels=response_models,
            responseParameters=response_parameters,
        )

        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigateway.method_response", name=name
        )
    if (not before) or resource_updated:
        after = await hub.exec.aws.apigateway.method_response.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )
        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result

        resource_translated = after["ret"]
        result["new_state"] = resource_translated
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    Deletes an API Gateway Method Response.

    Args:
        hub:

        ctx:

        name(string):
            An Idem name of the resource.

        resource_id(string, optional):
            Idem Resource id, formatted as: rest_api_id-parent_resource_id-http_method-status_code. Defaults to None.
                Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

        idem_test_aws_apigateway_method_response:
          aws.apigateway.method_response.absent:
            - name: value
            - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        # if resource_id isn't specified, the resource is considered to be absent.
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.method_response", name=name
        )
        return result

    before_ret = await hub.exec.aws.apigateway.method_response.get(
        ctx,
        name=name,
        resource_id=resource_id,
    )

    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]

    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.method_response", name=name
        )
        return result

    elif ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigateway.method_response", name=name
        )
        return result

    else:
        rest_api_id, parent_resource_id, http_method, status_code = resource_id.split(
            "-"
        )
        result["old_state"] = before_ret["ret"]
        delete_ret = await hub.exec.boto3.client.apigateway.delete_method_response(
            ctx,
            restApiId=rest_api_id,
            resourceId=parent_resource_id,
            httpMethod=http_method,
            statusCode=status_code,
        )

        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.apigateway.method_response", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the API Gateway Method Responses associated with a specific Method.

    Returns a list of apigateway.method_response descriptions

    Returns:
        Dict[str, Any]


    Examples:

        .. code-block:: bash

            $ idem describe aws.apigateway.method_response

    """

    result = {}

    get_rest_apis_ret = await hub.exec.boto3.client.apigateway.get_rest_apis(ctx)
    if not get_rest_apis_ret["result"]:
        hub.log.debug(f"Could not get Rest Apis {get_rest_apis_ret['comment']}")
        return result

    for rest_api in get_rest_apis_ret["ret"]["items"]:
        rest_api_id = rest_api.get("id")

        get_resources_ret = await hub.exec.boto3.client.apigateway.get_resources(
            ctx, restApiId=rest_api_id
        )
        if not get_resources_ret["result"]:
            hub.log.debug(
                f"Could not get Resources {get_resources_ret['comment']}. Will skip this Resource."
            )
            continue

        if get_resources_ret["ret"]["items"] is not None:
            for resource in get_resources_ret["ret"]["items"]:
                parent_resource_id = resource.get("id")
                if resource.get("resourceMethods") is not None:
                    for resource_method in resource.get("resourceMethods"):
                        method = await hub.exec.boto3.client.apigateway.get_method(
                            ctx,
                            restApiId=rest_api_id,
                            resourceId=parent_resource_id,
                            httpMethod=resource_method,
                        )
                        if not method["result"]:
                            hub.log.debug(
                                f"Could not get Resource Method {method['comment']}. Will skip this Method."
                            )
                            continue

                        http_method = method["ret"]["httpMethod"]
                        for method_response in method["ret"]["methodResponses"]:
                            resource_id = f"{rest_api_id}-{parent_resource_id}-{http_method}-{method_response}"

                            method_response = (
                                await hub.exec.aws.apigateway.method_response.get(
                                    ctx,
                                    resource_id=resource_id,
                                )
                            )
                            if (
                                not method_response["result"]
                                or not method_response["ret"]
                            ):
                                hub.log.debug(
                                    f"Could not get Resource Method Response {method_response['comment']}. Will skip this Method Response."
                                )
                                continue

                            resource_translated = method_response["ret"]
                            result[resource_translated["resource_id"]] = {
                                "aws.apigateway.method_response.present": [
                                    {parameter_key: parameter_value}
                                    for parameter_key, parameter_value in resource_translated.items()
                                ]
                            }

    return result



import copy
from typing import Any
from typing import Dict


"""
Exec functions for AWS DynamoDB Table resources.
"""


async def get(hub, ctx, name, resource_id) -> Dict[str, Any]:
    r"""Get a DynamoDB Table resource from AWS.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str):
            The name of the Dynamodb table.

    Returns:
        Dict[str, Any]
    """
    result = dict(comment=[], ret=None, result=True)

    describe_ret = await hub.exec.boto3.client.dynamodb.describe_table(
        ctx, TableName=resource_id
    )
    if not describe_ret["result"]:
        if "ResourceNotFoundException" in str(describe_ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.dynamodb.table", name=name
                )
            )
            result["comment"] += list(describe_ret["comment"])
            return result
        result["result"] = False
        result["comment"] = describe_ret["comment"]
        return result

    table_resource = describe_ret.get("ret").get("Table")

    tags_ret = await hub.exec.boto3.client.dynamodb.list_tags_of_resource(
        ctx, ResourceArn=table_resource.get("TableArn")
    )
    if not tags_ret["result"]:
        result["result"] = False
        result["comment"] = tags_ret["comment"]
        return result
    table_resource.update(copy.deepcopy(tags_ret["ret"]))

    continuous_backups_ret = (
        await hub.exec.boto3.client.dynamodb.describe_continuous_backups(
            ctx, TableName=table_resource.get("TableName")
        )
    )
    if not continuous_backups_ret.get("result"):
        result["result"] = False
        result["comment"] = continuous_backups_ret["comment"]
        return result
    table_resource.update(copy.deepcopy(continuous_backups_ret.get("ret")))

    time_to_live_ret = await hub.exec.boto3.client.dynamodb.describe_time_to_live(
        ctx, TableName=table_resource.get("TableName")
    )
    if not time_to_live_ret.get("result"):
        result["result"] = False
        result["comment"] = time_to_live_ret["comment"]
        return result
    table_resource.update(copy.deepcopy(time_to_live_ret.get("ret")))

    result["ret"] = hub.tool.aws.dynamodb.table.convert_raw_table_to_present(
        ctx, raw_resource=table_resource, idem_resource_name=name
    )

    return result



import boto3.session


def __init__(hub):
    # Create a single session for everything to be run from
    hub.tool.boto3.SESSION = boto3.session.Session()


def get(hub, botocore_session=None) -> boto3.session.Session:
    """
    Get the current boto3 session.
    """
    if botocore_session is None:
        # Reset the session if need be for thread safety
        hub.tool.boto3.SESSION = boto3.session.Session()
    return hub.tool.boto3.SESSION


def kwargs(hub, ctx):
    """
    Return the kwargs for a session object from ctx.acct in a consistent way
    """
    return dict(
        region_name=hub.tool.boto3.region.get(ctx),
        api_version=ctx.acct.get("api_version"),
        use_ssl=ctx.acct.get("use_ssl", True),
        endpoint_url=ctx.acct.get("endpoint_url"),
        aws_access_key_id=ctx.acct.get("aws_access_key_id"),
        aws_secret_access_key=ctx.acct.get("aws_secret_access_key"),
        aws_session_token=ctx.acct.get("aws_session_token"),
        verify=ctx.acct.get("verify"),
    )



"""Exec functions for s3 bucket's lifecycle configurations."""
from typing import Dict


async def get(hub, ctx, name: str, resource_id: str) -> Dict:
    """Returns the lifecycle configuration of a bucket.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str):
            AWS S3 bucket name.
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.s3.get_bucket_lifecycle_configuration(
        ctx, Bucket=resource_id
    )
    if not ret["result"]:
        if "NoSuchLifecycleConfiguration" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.s3.bucket_lifecycle", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    if "Rules" not in ret["ret"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.s3.bucket_lifecycle", name=name
            )
        )
        result["comment"] += list(ret["comment"])
        return result

    result[
        "ret"
    ] = hub.tool.aws.s3.bucket_lifecycle.convert_raw_bucket_lifecycle_to_present(
        bucket=resource_id, raw_resource=ret["ret"], idem_resource_name=name
    )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

"""
Util functions for AWS API Gateway v2 Domain Name resources.
"""


async def update(
    hub,
    ctx,
    resource_id: str,
    raw_resource: Dict[str, Any],
    resource_parameters: Dict[str, None],
) -> Dict[str, Any]:
    r"""
    Updates an AWS API Gateway v2 Domain Name resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        resource_id(string): The Domain Name resource identifier in Amazon Web Services.
        raw_resource(Dict): Existing resource parameters in Amazon Web Services.
        resource_parameters(Dict): Parameters from SLS file.

    Returns:
        Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)

    parameters = OrderedDict(
        {
            "domain_name_configurations": "DomainNameConfigurations",
            "mutual_tls_authentication": "MutualTlsAuthentication",
        }
    )

    parameters_to_update = {}

    domain_name_configurations = resource_parameters.get("domain_name_configurations")
    if domain_name_configurations is not None:
        if not hub.tool.aws.apigatewayv2.domain_name.are_domain_name_configurations_identical(
            domain_name_configurations,
            raw_resource.get("domain_name_configurations"),
        ):
            parameters_to_update[
                "DomainNameConfigurations"
            ] = domain_name_configurations

        resource_parameters.pop("domain_name_configurations")

    for key, value in resource_parameters.items():
        if value is not None and value != raw_resource.get(key):
            parameters_to_update[parameters[key]] = resource_parameters[key]

    if parameters_to_update:
        result["ret"] = {}
        for parameter_present, parameter_raw in parameters.items():
            if parameter_raw in parameters_to_update:
                result["ret"][parameter_present] = parameters_to_update[parameter_raw]

        if ctx.get("test", False):
            result["comment"] = (
                f"Would update parameters: " + ",".join(result["ret"].keys()),
            )
        else:
            update_ret = await hub.exec.boto3.client.apigatewayv2.update_domain_name(
                ctx,
                DomainName=resource_id,
                **parameters_to_update,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result

            result["comment"] = (
                f"Updated parameters: " + ",".join(result["ret"].keys()),
            )

    return result


def convert_raw_domain_name_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    r"""
    Convert AWS API Gateway v2 Domain Name resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        raw_resource(Dict[str, Any]): The AWS response to convert.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_parameters = OrderedDict(
        {
            "ApiMappingSelectionExpression": "api_mapping_selection_expression",
            "DomainNameConfigurations": "domain_name_configurations",
            "MutualTlsAuthentication": "mutual_tls_authentication",
            "Tags": "tags",
        }
    )
    resource_translated = {
        "resource_id": raw_resource.get("DomainName"),
        "name": raw_resource.get("DomainName"),
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


def are_domain_name_configurations_identical(
    hub,
    new_configuration: List,
    old_configuration: List,
) -> bool:
    r"""
    Compares the new and old domain name configurations.

    Args:
        hub: required for functions in hub.
        new_configuration(List): The new domain name configuration parameters.
        old_configuration(List): The old domain name configuration parameters.

    Returns:
        bool: true if there are no differences between the new and old domain name configurations.
    """

    if (new_configuration is None or len(new_configuration) == 0) and (
        old_configuration is None or len(old_configuration) == 0
    ):
        return True
    if (
        new_configuration is None
        or len(new_configuration) == 0
        or old_configuration is None
        or len(old_configuration) == 0
    ):
        return False

    diff = [
        i
        for i in new_configuration + old_configuration
        if i not in new_configuration or i not in old_configuration
    ]

    return len(diff) == 0



import copy
from typing import Any
from typing import Dict


async def update_tags(
    hub, ctx, old_tags: Dict[str, Any], new_tags: Dict[str, Any], resource_id: str
):
    """
    Update tags of AWS auto scaling group
    Args:
        hub:
        ctx:
        old_tags: dict of old tags in the format of {tag-key: tag-value, propagate_at_launch-tag_key: value}
        new_tags: dict of new tags in the format of {tag-key: tag-value, propagate_at_launch-tag_key: value}
        resource_id: The name of the Auto Scaling group.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": dict of updated tags}

    """
    tags_to_add = {}
    tags_to_remove = {}
    if new_tags is not None:
        (
            tags_to_remove,
            tags_to_add,
        ) = hub.tool.aws.tag_utils.diff_auto_scaling_dict_tags(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            list_tags_to_remove = (
                hub.tool.aws.tag_utils.convert_auto_scaling_tag_dict_to_list(
                    tags_to_remove
                )
            )
            raw_tags_to_remove = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_present_tags_to_raw_tags(
                resource_id, list_tags_to_remove
            )
            delete_ret = await hub.exec.boto3.client.autoscaling.delete_tags(
                ctx, Tags=raw_tags_to_remove
            )
            if not delete_ret["result"]:
                result["comment"] = (delete_ret["comment"],)
                result["result"] = False
                return result

    if tags_to_add:
        if not ctx.get("test", False):
            list_tags_to_add = (
                hub.tool.aws.tag_utils.convert_auto_scaling_tag_dict_to_list(
                    tags_to_add
                )
            )
            raw_tags_to_add = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_present_tags_to_raw_tags(
                resource_id, list_tags_to_add
            )
            add_ret = await hub.exec.boto3.client.autoscaling.create_or_update_tags(
                ctx, Tags=raw_tags_to_add
            )
            if not add_ret["result"]:
                result["comment"] = (add_ret["comment"],)
                result["result"] = False
                return result
    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result



import time
from typing import Any
from typing import Dict

"""
Function for getting the SQS queue state via its URL
"""


async def get(
    hub,
    ctx,
    resource_id: str,
    expected_attributes: Dict = None,
    expected_tags: Dict = None,
    max_retries: int = 45,
    delay_between_retries: int = 1,
) -> Dict[str, Any]:
    """
    Returns the state of an SQS queue

    Args:
        hub: required for functions in hub
        ctx(Dict):
        resource_id(string): the URL of the queue
        expected_attributes(Dict, optional): retries getting the attributes until they contain these expected attributes
        expected_tags(Dict, optional): retries getting the tags until they contain these expected tags
        max_retries(int, optional): max retries for getting the attributes and/or the tags
            when the expected_attributes and/or expected_tags are set
        delay_between_retries(int, optional): the delay in seconds between the retries

    Returns: The state of the queue which consists of its
        idem resource name, queue name, resource_id, attributes and tags
    """

    result = dict(ret=None, comment=(), result=True)
    name = resource_id.split("/")[-1]
    ret = dict(name=name, resource_id=resource_id)
    new_queue_attributes_ret = {}

    if expected_attributes:

        # Poll get_queue_attributes() until it returns the expected attributes
        for i in range(max_retries):
            # Get the queue attributes
            new_queue_attributes_ret = (
                await hub.exec.boto3.client.sqs.get_queue_attributes(
                    ctx, QueueUrl=resource_id, AttributeNames=["All"]
                )
            )

            if new_queue_attributes_ret["result"]:
                present_attributes = (
                    hub.tool.aws.sqs.conversion_utils.convert_raw_attributes_to_present(
                        new_queue_attributes_ret["ret"].get("Attributes")
                    )
                )

                if hub.tool.aws.sqs.queue_utils.compare_present_queue_attributes(
                    expected_attributes, present_attributes
                ):
                    break

            time.sleep(delay_between_retries)
    else:
        # Get the queue attributes
        new_queue_attributes_ret = await hub.exec.boto3.client.sqs.get_queue_attributes(
            ctx, QueueUrl=resource_id, AttributeNames=["All"]
        )

    if new_queue_attributes_ret["result"]:
        # Convert the raw queue attributes to present form and add them to the final result
        present_attributes = (
            hub.tool.aws.sqs.conversion_utils.convert_raw_attributes_to_present(
                new_queue_attributes_ret["ret"].get("Attributes")
            )
        )
        ret.update(present_attributes.items())
    else:
        error_comment = new_queue_attributes_ret["comment"]
        if "QueueDoesNotExist" in str(error_comment):
            result["comment"] += (
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.sqs.queue", name=name
                ),
            )
        else:
            result["result"] = False
        result["comment"] += error_comment
        return result

    new_queue_tags_ret = {}
    if expected_tags:

        # Poll list_queue_tags() until it returns the expected tags
        for i in range(max_retries):
            # Get the queue tags
            new_queue_tags_ret = await hub.exec.boto3.client.sqs.list_queue_tags(
                ctx, QueueUrl=resource_id
            )

            if new_queue_tags_ret["result"]:
                if expected_tags.items() <= new_queue_tags_ret["ret"]["Tags"].items():
                    break

            time.sleep(delay_between_retries)
    else:
        # Get the queue tags
        new_queue_tags_ret = await hub.exec.boto3.client.sqs.list_queue_tags(
            ctx, QueueUrl=resource_id
        )

    if new_queue_tags_ret["result"]:
        # Add the queue tags to the final result
        ret["tags"] = dict(new_queue_tags_ret["ret"].get("Tags", {}))
    else:
        error_comment = new_queue_tags_ret["comment"]
        if "QueueDoesNotExist" in str(error_comment):
            result["comment"] += (
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.sqs.queue", name=name
                ),
            )
        else:
            result["result"] = False
        result["comment"] += error_comment
        return result

    result["ret"] = ret

    return result



from typing import Any
from typing import Dict


def sanitize_distribution_config(hub, original_distribution_config: Dict[str, Any]):
    updated_distribution_config = {}

    for (
        distribution_option_key,
        distribution_option_value,
    ) in original_distribution_config.items():
        if distribution_option_value is not None:
            updated_distribution_config[
                distribution_option_key
            ] = distribution_option_value

    return updated_distribution_config


def sanitize_distribution_config_for_update(
    hub, original_distribution_config: Dict[str, Any]
):
    updated_distribution_config = {}
    mandatory_params_in_update_with_defaults = {
        "Aliases": {"Quantity": 0},
        "DefaultRootObject": "",
        "CacheBehaviors": {"Quantity": 0},
        "CustomErrorResponses": {"Quantity": 0},
        "Logging": {
            "Enabled": False,
            "IncludeCookies": False,
            "Bucket": "",
            "Prefix": "",
        },
        "PriceClass": "PriceClass_All",
        "ViewerCertificate": {
            "CertificateSource": "cloudfront",
            "CloudFrontDefaultCertificate": True,
            "MinimumProtocolVersion": "TLSv1",
            "SSLSupportMethod": "vip",
        },
        "Restrictions": {"GeoRestriction": {"RestrictionType": "none", "Quantity": 0}},
        "WebACLId": "",
        "HttpVersion": "http2",
    }
    for (
        distribution_option_key,
        distribution_option_value,
    ) in original_distribution_config.items():
        if distribution_option_value is not None:
            updated_distribution_config[
                distribution_option_key
            ] = distribution_option_value
        elif distribution_option_key in mandatory_params_in_update_with_defaults:
            updated_distribution_config[
                distribution_option_key
            ] = mandatory_params_in_update_with_defaults[distribution_option_key]

    return updated_distribution_config



from typing import List


async def apply(
    hub, ctx, resource, *, old_value: str, new_value: str, comments: List[str]
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    ret = await hub.exec.boto3.client.ec2.associate_address(
        ctx,
        InstanceId=resource.id,
        PrivateIpAddress=new_value,
    )
    if ret.comment:
        comments.append(ret.comment)
    return ret.result



"""
hub.exec.boto3.client.ec2.allocate_address
hub.exec.boto3.client.ec2.describe_addresses
hub.exec.boto3.client.ec2.release_address
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]
TREQ = {
    "absent": {
        "require": ["aws.ec2.nat_gateway.absent"],
    }
}


async def present(
    hub,
    ctx,
    name: str,
    domain: str = "standard",
    resource_id: str = None,
    network_border_group: str = None,
    public_ipv4_pool: str = None,
    customer_owned_ipv4_pool: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""Allocates an Elastic IP address to your Amazon Web Services account.

    After you allocate the Elastic IP address you can associate it with an instance or network interface. After you release an Elastic IP address,
    it is released to the IP address pool and can be allocated to a different Amazon Web Services account. [EC2-VPC]
    If you release an Elastic IP address, you might be able to recover it. You cannot recover an Elastic IP address
    that you released after it is allocated to another Amazon Web Services account. You cannot recover an Elastic IP
    address for EC2-Classic. To attempt to recover an Elastic IP address that you released, specify it in this
    operation. An Elastic IP address is for use either in the EC2-Classic platform or in a VPC. By default,
    you can allocate 5 Elastic IP addresses for EC2-Classic per Region and 5 Elastic IP addresses for EC2-VPC per
    Region. For more information, see Elastic IP Addresses in the Amazon Elastic Compute Cloud User Guide.

    Args:
        name(str):
            An Idem name of the elastic IP resource.

        domain(str):
            Indicates whether the Elastic IP address is for use with instances in a VPC or instances in EC2-Classic.

        resource_id(str, Optional):
            The public ip address.

        network_border_group(str, Optional):
            A unique set of Availability Zones, Local Zones, or Wavelength Zones from which Amazon Web Services
            advertises IP addresses.

        public_ipv4_pool(str, Optional):
            The ID of an address pool that you own.

        customer_owned_ipv4_pool(str, Optional):
            The ID of a customer-owned address pool.

        tags(dict or list, Optional):
            Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the elastic IP resource. Defaults to None.

            * Key (str):
                The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not
                begin with aws: .

            * Value (str):
                The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.

    Request Syntax:
        .. code-block:: sls

           [elastic_ip-resource-id]:
             aws.ec2.elastic_ip.present:
               - name: "string"
               - domain: "string"
               - tags:
                   scope: "string"
                   name: "string"

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

           127.15.17.187:
             aws.ec2.elastic_ip.present:
               - name: 127.15.17.187
               - resource_id: 127.15.17.187
               - domain: vpc
               - tags:
                   - Key: name
                     Value: ONE
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    allocation_id = None
    public_ip = None
    before = None
    resource_updated = False
    try:
        if resource_id:
            before = await hub.exec.aws.ec2.elastic_ip.get(
                ctx, name=name, resource_id=resource_id
            )

        if isinstance(tags, List):
            tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

        if before:
            if not before["result"] or not before["ret"]:
                result["result"] = False
                result["comment"] = before["comment"]
                return result

            public_ip = resource_id
            result["old_state"] = copy.deepcopy(before["ret"])
            plan_state = copy.deepcopy(result["old_state"])
            if domain == "vpc":
                allocation_id = before["ret"]["allocation_id"]
            old_tags = result["old_state"].get("tags")
            if tags is not None and tags != old_tags:
                update_ret = await hub.exec.aws.ec2.tag.update_tags(
                    ctx=ctx,
                    resource_id=allocation_id,
                    old_tags=old_tags,
                    new_tags=tags,
                )
                if not update_ret["result"]:
                    result["comment"] = update_ret["comment"]
                    result["result"] = False
                    return result

                if update_ret["ret"]:
                    result["comment"] = update_ret["comment"]
                resource_updated = resource_updated or bool(update_ret["result"])
                if ctx.get("test", False) and update_ret["result"]:
                    plan_state["tags"] = update_ret["ret"]

        else:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "domain": domain,
                        "network_border_group": network_border_group,
                        "public_ipv4_pool": public_ipv4_pool,
                        "customer_owned_ipv4_pool": customer_owned_ipv4_pool,
                        "allocation_id": "allocation_id_known_after_present",
                        "tags": tags,
                    },
                )
                result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.ec2.elastic_ip", name=name
                )
                return result

            ret = await hub.exec.boto3.client.ec2.allocate_address(
                ctx,
                Domain=domain,
                PublicIpv4Pool=public_ipv4_pool,
                NetworkBorderGroup=network_border_group,
                CustomerOwnedIpv4Pool=customer_owned_ipv4_pool,
                TagSpecifications=[
                    {
                        "ResourceType": "elastic-ip",
                        "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                    }
                ]
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["result"] = ret["result"]
                result["comment"] = ret["comment"]
                return result
            public_ip = ret["ret"]["PublicIp"]
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.ec2.elastic_ip", name=name
            )
    except Exception as e:
        result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
        result["result"] = False
        return result

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.aws.ec2.elastic_ip.get(
                ctx, name=name, resource_id=public_ip
            )
            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""Releases the specified Elastic IP address.

    [EC2-Classic, default VPC] Releasing an Elastic IP address automatically disassociates it from any instance that
    it's associated with. To disassociate an Elastic IP address without releasing it, use DisassociateAddress .
    [Nondefault VPC] You must use DisassociateAddress to disassociate the Elastic IP address before you can release it.
    After releasing an Elastic IP address, it is released to the IP address pool. Be sure to update your DNS records and
    any servers or devices that communicate with the address. If you attempt to release an Elastic IP address that you
    already released, you'll get an AuthFailure error if the address is already allocated to another Amazon Web Services
    account. [EC2-VPC] After you release an Elastic IP address for use in a VPC, you might be able to recover it. For
    more information, see AllocateAddress.

    Args:
        name(str): An Idem name of the elastic IP resource.

        resource_id(str, Optional): The public ip address. Idem automatically considers this resource being absent
         if this field is not specified.


    Request Syntax:
        .. code-block:: sls

           [elastic_ip-resource-id]:
             aws.ec2.elastic_ip.absent:
               - name: "string"
               - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            127.15.17.187:
              aws.ec2.elastic_ip.absent:
                - name: 127.15.17.187
                - resource_id: 127.15.17.187
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.elastic_ip", name=name
        )

        return result
    before = await hub.exec.aws.ec2.elastic_ip.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before and not before["result"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.elastic_ip", name=name
        )
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.elastic_ip", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ec2.elastic_ip", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        domain = result["old_state"]["domain"]
        try:
            if domain == "vpc":
                allocation_id = result["old_state"]["allocation_id"]
                ret = await hub.exec.boto3.client.ec2.release_address(
                    ctx, AllocationId=allocation_id
                )
            else:
                ret = await hub.exec.boto3.client.ec2.release_address(
                    ctx, PublicIp=resource_id
                )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.ec2.elastic_ip", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Describes the specified Elastic IP addresses or all of your Elastic IP addresses.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ec2.elastic_ip
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_addresses(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe elastic_ip {ret['comment']}")
        return {}

    for each_elastic_ip in ret["ret"]["Addresses"]:
        resource_id = each_elastic_ip.get("PublicIp")
        resource_translated = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_elastic_ip_to_present(
                raw_resource=each_elastic_ip,
                idem_resource_name=resource_id,
            )
        )

        result[resource_id] = {
            "aws.ec2.elastic_ip.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



"""Exec module for managing Amazon Cloudfront Distribution."""
from typing import Dict


async def get(
    hub,
    ctx,
    name,
    resource_id: str,
) -> Dict:
    """Use an un-managed Cloudfront distribution as a data-source. Supply resource_id as a filter.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str):
            AWS cloudfront distribution id to identify the resource.

    Returns:
        .. code-block:: python

            {"result": True|False, "comment": A message List, "ret": None|Dict}

    Examples:
        Calling this exec module function from the cli with name and resource_id

        .. code-block:: bash

            idem exec aws.cloudfront.distribution.get name="name" resource_id="resource_id"

        Calling this exec module function from within a state module in pure python.

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id):
                before = await hub.exec.aws.cloudfront.distribution.get(
                ctx, name=name, resource_id=resource_id
            )
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.cloudfront.get_distribution(ctx, Id=resource_id)
    if not ret["result"]:
        if "NoSuchDistribution" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.cloudfront.distribution", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Distribution"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.cloudfront.distribution", name=name
            )
        )
        return result
    arn = ret["ret"]["Distribution"]["ARN"]
    before_tag = await hub.exec.boto3.client.cloudfront.list_tags_for_resource(
        ctx, Resource=arn
    )
    if not before_tag["result"]:
        result["result"] = False
        result["comment"] = before_tag["comment"]
        return result

    if before_tag["ret"].get("Tags"):
        ret["ret"]["Distribution"]["Tags"] = (
            before_tag["ret"].get("Tags").get("Items", [])
        )
    ret["ret"]["Distribution"]["ETag"] = ret["ret"]["ETag"]
    result[
        "ret"
    ] = await hub.tool.aws.cloudfront.conversion_utils.convert_raw_distribution_to_present(
        ctx, raw_resource=ret["ret"]["Distribution"], idem_resource_name=name
    )
    return result



"""Exec module for managing Amazon Cloudwatchlogs Resource Policy."""
from typing import Any
from typing import Dict


async def get(
    hub,
    ctx,
    name: str,
) -> Dict[str, Any]:
    r"""Returns resource policy object for the given resource policy name.

    Args:
        name(str): Name of the new policy. An Idem name of the resource

    Returns:
        .. code-block:: python

            {"result": True|False, "comment": A message List, "ret": None|Dict}

    Examples:
        Calling this exec module function from the cli with name

        .. code-block:: bash

            idem exec aws.cloudwatchlogs.resource_policy.get name="name"

        Calling this exec module function from within a state module in pure python.

        .. code-block:: python

            async def state_function(hub, ctx, resource_id):
                before = await hub.exec.aws.cloudwatchlogs.resource_policy.get(
                ctx, name=resource_id
            )
    """
    result = dict(comment=(), result=True, ret=None)
    ret = await hub.exec.boto3.client.logs.describe_resource_policies(ctx)
    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"] = result["comment"] + ret["comment"]
        return result
    if ret["ret"]:
        resource_policies = ret["ret"]["resourcePolicies"]
        for resource_policy in resource_policies:
            if resource_policy.get("policyName") == name:
                result["ret"] = resource_policy
                break
    return result



"""Exec module for managing Amazon Application Autoscaling scalable target."""
from typing import Dict
from typing import List


async def get(
    hub,
    ctx,
    name,
    service_namespace: str,
    scaling_resource_ids: List[str] = None,
    scalable_dimension: str = None,
) -> Dict:
    """Use an un-managed scaling target as a data-source. Supply one of the inputs as the filter.

    Args:
        name(str):
            The name of the Idem state.

        service_namespace(str):
            The namespace of the Amazon Web Services service that provides the resource.
            For a resource provided by your own application or service, use custom-resource instead.

        scalable_dimension(str):
            The scalable dimension. This string consists of the service namespace, resource type,
            and scaling property. ecs:service:DesiredCount

        scaling_resource_ids(list[str], Optional):
            The identifier of the resource associated with the scaling target.
            This string consists of the resource type and unique identifier.

    Returns:
        .. code-block:: python

            {"result": True|False, "comment": A message List, "ret": None|Dict}

    Examples:
        Calling this exec module function from the cli with service_namespace

        .. code-block:: bash

            idem exec aws.application_autoscaling.scalable_target.get service_namespace="service_namespace"

        Calling this exec module function from within a state module in pure python.

        .. code-block:: python

            async def state_function(hub, ctx, name, service_namespace, scaling_resource_id, scalable_dimension):
                before = await hub.exec.aws.application_autoscaling.scalable_target.get(
                ctx=ctx,
                name=name,
                service_namespace=service_namespace,
                scaling_resource_ids=[scaling_resource_id],
                scalable_dimension=scalable_dimension,
            )
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client[
        "application-autoscaling"
    ].describe_scalable_targets(
        ctx=ctx,
        ServiceNamespace=service_namespace,
        ResourceIds=scaling_resource_ids,
        ScalableDimension=scalable_dimension,
    )
    if not ret["result"]:
        result["comment"] = ret["comment"]
        result["result"] = False
        return result
    if not ret["ret"]["ScalableTargets"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.application_autoscaling.scaling_target", name=name
            )
        )
        return result

    resource = ret["ret"]["ScalableTargets"][0]
    if len(ret["ret"]["ScalableTargets"]) > 1:
        result["comment"].append(
            f"More than one aws.application_autoscaling.scaling_target resource was found. Use resource {resource.get('ResourceId')}"
        )
    result[
        "ret"
    ] = hub.tool.aws.application_autoscaling.conversion_utils.convert_raw_scaling_target_to_present(
        ctx, raw_resource=resource, idem_resource_name=name
    )
    return result



"""State module for managing IAM Groups."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "absent": {"require": ["aws.iam.user.absent"]},
}


async def present(
    hub, ctx, name: str, group_name: str, resource_id: str = None, path: str = None
) -> Dict[str, Any]:
    """Create/Update the IAM Group.

    Creates a new group.  For information about the number of groups you can create, see IAM and STS quotas in the
    IAM User Guide.

    Args:
        name(str): An Idem name of the resource. This is also used as the name of the IAM group.

        resource_id(str, optional): An identifier of the resource in the provider. Name of Group.

        path(str, optional): The path to the group. This parameter is optional. If it is not included,
            it defaults to a slash (/). This parameter allows (through its regex pattern) a string of
            characters consisting of either a forward slash (/) by itself or a string that must begin and
            end with forward slashes. For more information about paths, see IAM identifiers in the IAM User Guide.

        group_name(str): The name of the group to create. Do not include the path in this value.
            IAM user, group, role, and policy names must be unique within the account. Names are not distinguished by
            case. For example, you cannot create resources named both "MyResource" and "myresource".

    Request Syntax:
        .. code-block:: sls

            [group_name]:
              aws.iam.group.present:
                - name: "string"
                - group_name: "string"
                - resource_id: "string"
                - path: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            test-group:
              aws.iam.group.present:
                - name: test-group
                - resource_id: test-group
                - group_name: test-group
                - path: /

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.iam.group.get(
            ctx, resource_id=resource_id, name=name
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["comment"] += hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.iam.group", name=name
        )
        result["old_state"] = copy.deepcopy(before["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])

        update_ret = await hub.tool.aws.iam.group.update(
            ctx,
            resource_id=resource_id,
            before=before["ret"],
            group_name=group_name,
            path=path,
        )
        result["comment"] += update_ret["comment"]
        if not update_ret["result"]:
            result["result"] = False
            return result

        if update_ret["ret"] and "group_name" in update_ret["ret"]:
            resource_id = group_name
        resource_updated = bool(update_ret["ret"])

        if update_ret["ret"] and ctx.get("test", False):
            result["new_state"].update(update_ret["ret"])
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.iam.group", name=name
            )
        if resource_updated and ctx.get("test", False):
            return result

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": name,
                    "group_name": group_name,
                    "path": path,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.iam.group", name=name
            )
            return result
        create_ret = await hub.exec.boto3.client.iam.create_group(
            ctx, GroupName=group_name, Path=path
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.iam.group", name=name
        )
        resource_id = group_name

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.aws.iam.group.get(
            ctx, name=group_name, resource_id=resource_id
        )
        if not after_ret["result"] or not after_ret["ret"]:
            result["result"] = False
            result["comment"] += tuple(after_ret["comment"])
            return result
        result["new_state"] = after_ret["ret"]

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes the specified IAM group.

    The group must not contain any users or have any attached policies.

    Args:
        name(str): An Idem name of the resource.
        resource_id(str): Name of IAM group.

    Request Syntax:
        .. code-block:: sls

            [group_name]:
              aws.iam.group.absent:
                - name: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            test-group:
              aws.iam.group.absent:
                - name: test-group
                - resource_id: test-group
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.group", name=name
        )
        return result

    before = await hub.exec.aws.iam.group.get(ctx, name=name, resource_id=resource_id)
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.group", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.iam.group", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        delete_ret = await hub.exec.boto3.client.iam.delete_group(
            ctx, GroupName=resource_id
        )
        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.iam.group", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Gets information about the IAM Group.

    Describe the resource in a way that can be recreated/managed with the corresponding
    present function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.iam.group

    """
    result = {}
    ret = await hub.exec.boto3.client.iam.list_groups(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe iam group {ret['comment']}")
        return {}

    for group in ret["ret"]["Groups"]:
        group_name = group.get("GroupName")

        translated_resource = (
            hub.tool.aws.iam.conversion_utils.convert_raw_group_to_present(
                raw_resource=group,
                idem_resource_name=group_name,
            )
        )
        resource_key = f"iam-group-{group_name}"
        result[resource_key] = {
            "aws.iam.group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



"""State module for managing AWS CloudWatch Log Groups."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    kms_key_id: str = None,
    tags: Dict[str, str] = None,
    retention_in_days: int = None,
) -> Dict[str, Any]:
    r"""Creates a log group with the specified name.

    You can create up to 5000 log groups per account.

    You must use the following guidelines when naming a log group:
      - Log group names must be unique within a region for an AWS account.
      - Log group names can be between 1 and 512 characters long.
      - Log group names consist of the following characters: a-z, A-Z, 0-9, '_' (underscore), '-' (hyphen),
        '/' (forward slash), and '.' (period).

    If you associate a AWS Key Management Service (AWS KMS) customer master key (CMK) with the log group, ingested data
    is encrypted using the CMK. This association is stored as long as the data encrypted with the CMK is still within
    Amazon CloudWatch Logs. This enables Amazon CloudWatch Logs to decrypt this data whenever it is requested.

    If you attempt to associate a CMK with the log group but the CMK does not exist or the CMK is disabled, you will
    receive an InvalidParameterException error.

    Args:
        name(str):
            An Idem name of the log group.
        resource_id(str, Optional):
            AWS CloudWatch Log Group Name Prefix.
        kms_key_id(str, Optional):
            The Amazon Resource Name (ARN) of the CMK to use when encrypting log data.
        tags(dict, Optional):
            The key-value pairs to use for the tags.
            CloudWatch Logs doesn’t support IAM policies that prevent users from assigning specified tags to log groups
            using the aws:Resource/key-name  or aws:TagKeys condition keys. For more information about using tags to
            control access, see Controlling access to Amazon Web Services resources using tags.
            Defaults to None.
        retention_in_days(int, Optional):
            The number of days to retain the log events in the specified log group.
            Possible values are: 1, 3, 5, 7, 14, 30, 60, 90, 120, 150, 180, 365, 400, 545, 731, 1827, and 3653.

    Request Syntax:
      .. code-block:: sls

        [log_group_name]:
          aws.cloudwatch.log_group.present:
            - name: "string"
            - kms_key_id: "string"
            - tags:
                - "string": "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.cloudwatch.log_group.present:
                - name: value
                - tags:
                     TestKey4: TestValue4
                - kms_key_id: "arn:aws:kms:us-east-1:xxxxxxxxx:key/xxxx-9500-xxxxx"
    """
    result = dict(comment=(), name=name, old_state=None, new_state=None, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.cloudwatch.log_group.get(
            ctx, name=name, resource_id=resource_id
        )
    try:
        if before and before["result"] and before["ret"]:
            result["old_state"] = copy.deepcopy(before["ret"])
            plan_state = copy.deepcopy(result["old_state"])
            result["comment"] = (f"aws.cloudwatch.log_group '{name}' already present.",)
            if tags is not None and result["old_state"].get("tags") != tags:
                update_ret = await hub.tool.aws.cloudwatch.tag.update_tags(
                    ctx=ctx,
                    resource_name=resource_id,
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])
                if ctx.get("test", False) and update_ret["result"]:
                    plan_state["tags"] = update_ret["ret"]
            if retention_in_days and (
                "retention_in_days" not in result["old_state"]
                or result["old_state"]["retention_in_days"] != retention_in_days
            ):
                if ctx.get("test", False):
                    plan_state["retention_in_days"] = retention_in_days
                else:
                    ret = await hub.exec.boto3.client.logs.put_retention_policy(
                        ctx, logGroupName=name, retentionInDays=retention_in_days
                    )
                    result["comment"] = result["comment"] + (ret["comment"])
                    result["result"] = ret["result"]
                    resource_updated = resource_updated or bool(ret["ret"])
        else:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "kms_key_id": kms_key_id,
                        "arn": "arn_known_after_present",
                        "tags": tags,
                        "retention_in_days": retention_in_days,
                    },
                )
                result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.cloudwatch.log_group", name=name
                )
                return result
            ret = await hub.exec.boto3.client.logs.create_log_group(
                ctx, logGroupName=name, kmsKeyId=kms_key_id, tags=tags
            )
            resource_id = name
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.cloudwatch.log_group", name=name
            )
            if retention_in_days:
                ret = await hub.exec.boto3.client.logs.put_retention_policy(
                    ctx, logGroupName=name, retentionInDays=retention_in_days
                )
                result["result"] = ret["result"]
                result["comment"] = result["comment"] + (ret["comment"])
                if not result["result"]:
                    return result
    except hub.tool.boto3.exception.ClientError as e:
        result["result"] = False
        result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.aws.cloudwatch.log_group.get(
                ctx, name=name, resource_id=resource_id
            )
            if after and after["result"] and after["ret"]:
                result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""Deletes the specified cloudwatch log group.

    Args:
        name(str): An Idem name of the Log Group.
        resource_id(str, Optional): AWS CloudWatch Log Group Name Prefix. Idem automatically considers this resource
         being absent if this field is not specified.

    Request Syntax:
        .. code-block:: sls

            [log_group_name]:
              aws.cloudwatch.log_group.absent:
                - name: value
                - resource_id: value

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.cloudwatch.log_group.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), name=name, old_state=None, new_state=None, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.cloudwatch.log_group", name=name
        )
        return result
    before = await hub.exec.aws.cloudwatch.log_group.get(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.cloudwatch.log_group", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.cloudwatch.log_group", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.logs.delete_log_group(
                ctx, logGroupName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.cloudwatch.log_group", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Lists the AWS CloudWatch Log Groups.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.cloudwatch.log_group
    """
    result = {}
    ret = await hub.exec.boto3.client.logs.describe_log_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe log groups {ret['comment']}")
        return {}

    for resource in ret["ret"]["logGroups"]:
        resource_id = resource.get("logGroupName")
        resource_translated = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_log_group_to_present(
            ctx, raw_resource=resource, idem_resource_name=resource_id
        )
        result[resource_id] = {
            "aws.cloudwatch.log_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from typing import Dict

__contracts__ = ["soft_fail"]


def build(
    hub,
    ctx,
    service: str,
    resource: str,
    partition: str = "aws",
    region: str = None,
    account_id: str = None,
) -> Dict:
    """
    Returns the ARN string from parameters.

    Args:
        hub: The redistributed pop central hub.
        ctx: The context.
        service(str): The service namespace that identifies the AWS product (for example, Amazon S3, IAM, or Amazon RDS). For a list of
                namespaces, see http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#genref-aws-service-namespaces.
        resource(str): The content of this part of the ARN varies by service. It often includes an indicator of the type of resource —
                for example, an IAM user or Amazon RDS database - followed by a slash (/) or a colon (:), followed by the
                resource name itself. Some services allows paths for resource names, as described in
                http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#arns-paths.
        partition(str, optional): The partition that the resource is in. For standard AWS regions, the partition is "aws" (default).
                If you have resources in other partitions, the partition is "aws-partitionname". For example, the partition for resources
                in the China (Beijing) region is "aws-cn".
        region(str, optional): The region the resource resides in. Note that the ARNs for some resources do not require a region, so this
                component might be omitted.
        account_id(str, optional): The ID of the AWS account that owns the resource, without the hyphens. For example, 123456789012. Note that the
                ARNs for some resources don't require an account number, so this component might be omitted.

    .. code-block:: bash

        $ idem exec aws.arn.build service=ec2 region=us-west-2 resource=my-resource
    """
    ret = dict(comment=[], result=True, ret=None)

    arn = hub.tool.aws.arn_utils.build(
        partition=partition,
        service=service,
        region=region,
        account_id=account_id,
        resource=resource,
    )
    ret["ret"] = arn

    return ret


def parse(hub, ctx, arn: str) -> Dict:
    """
    Parses the provided AWS ARN and returns its components.

    Args:
        hub: The redistributed pop central hub.
        ctx: The context.
        arn(string): The AWS ARN.

    .. code-block:: bash

        $ idem exec aws.arn.parse arn=arn:aws:ec2:us-west-2:accountid:resource
    """
    ret = dict(comment=[], result=True, ret=None)

    try:
        parts = hub.tool.aws.arn_utils.parse(arn=arn)
    except TypeError as e:
        ret["result"] = False
        ret["comment"] = (str(e),)
        return ret

    ret["ret"] = parts

    return ret



"""State module for managing Amazon Organizations Policy Attachments."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.organizations.organization.present",
            "aws.organizations.organization_unit.present",
            "aws.organizations.account.present",
            "aws.organizations.policy.present",
        ],
    },
}


async def present(
    hub, ctx, name: str, policy_id: str, target_id: str
) -> Dict[str, Any]:
    """Attaches a policy to a root, an organizational unit (OU), or an individual account.

    How the policy affects accounts depends on the type of policy. The only supported policy types are
    ``SERVICE_CONTROL_POLICY`` and ``TAG_POLICY``.

    Args:
        name(str):
            An Idem name of the resource.
        policy_id(str):
            The unique identifier (ID) of the policy that you want to attach to the target.
        target_id(str):
            The unique identifier (ID) of the root, OU, or account that you want to attach the policy to.

    Request syntax:
      .. code-block:: sls

        [idem_test_aws_organizations_policy_attachment]:
          aws.organizations.policy_attachment.present:
            - name: 'string'
            - policy_id: 'string'
            - target_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_organizations_policy_attachment:
              aws.organizations.policy_attachment.present:
                - name: 'idem_test_policy_attachment'
                - policy_id: 'p-123456789012'
                - target_id: 'ou-root-id'
    """
    result = dict(comment="", name=name, result=True, old_state=None, new_state=None)

    try:
        before_ret = await hub.tool.aws.organizations.policy_attachment.is_target_policy_attached(
            ctx, policy_id=policy_id, target_id=target_id
        )
    except hub.tool.boto3.exception.ClientError as e:
        result["comment"] = f"{e.__class__.__name__}: {e}"
        result["result"] = False
        return result

    # If before_ret has False result and has an error message in comment, then immediately return with the error.
    if before_ret["result"] is False and before_ret["comment"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    if ctx.get("test", False):
        if before_ret["result"]:
            result[
                "comment"
            ] = f"aws.organizations.policy_attachment {name} already exists"
        else:
            result[
                "comment"
            ] = f"Would attach aws.organizations.policy_attachment {name}"
        return result

    if before_ret["result"]:
        result["old_state"] = before_ret["ret"]
        result["new_state"] = copy.deepcopy(before_ret["ret"])
        result["comment"] = f"'{name}' already exists"
    else:
        try:
            ret = await hub.exec.boto3.client.organizations.attach_policy(
                ctx,
                PolicyId=policy_id,
                TargetId=target_id,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Attached '{name}'"
            result["new_state"] = {"PolicyId": policy_id}
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
    return result


async def absent(hub, ctx, name: str, policy_id: str, target_id: str) -> Dict[str, Any]:
    """Detaches a policy from a target root, organizational unit (OU), or account.

    .. warning::
        If the policy being detached is a service control policy (SCP), the changes to permissions for
        Identity and Access Management (IAM) users and roles in affected accounts are immediate.

    Every root, OU, and account must have at least one SCP attached. If you want to replace the default
    ``FullAWSAccess`` policy with an SCP that limits the permissions that can be delegated, you must attach
    the replacement SCP before you can remove the default SCP. This is the authorization strategy of an
    ``allow list``. If you instead attach a second SCP and leave the ``FullAWSAccess`` SCP still attached,
    and specify ``"Effect": "Deny"`` in the second SCP to override the ``"Effect": "Allow"`` in the
    ``FullAWSAccess`` policy (or any other attached SCP), you're using the authorization strategy of a ``deny list``.

    Args:
        name(str):
            An Idem name of the resource.
        policy_id(str):
            The unique identifier (ID) of the policy you want to detach.
        target_id(str):
            The unique identifier (ID) of the root, OU, or account that you want to detach the policy from.

    Request syntax:
      .. code-block:: sls

        [idem_test_aws_organizations_policy_attachment]:
          aws.organizations.policy_attachment.absent:
            - name: 'string'
            - policy_id: 'string'
            - target_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_organizations_policy_attachment:
              aws.organizations.policy_attachment.absent:
                - name: 'idem_test_policy_attachment'
                - policy_id: 'p-123456789012'
                - target_id: 'ou-root-id'
    """
    result = dict(comment="", name=name, result=True, old_state=None, new_state=None)

    try:
        before_ret = await hub.tool.aws.organizations.policy_attachment.is_target_policy_attached(
            ctx, policy_id=policy_id, target_id=target_id
        )
    except hub.tool.boto3.exception.ClientError as e:
        result["comment"] = f"{e.__class__.__name__}: {e}"
        result["result"] = False
        return result

    if before_ret["result"] is False and before_ret["comment"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    if not before_ret["result"]:
        result["comment"] = f"'{name}' already absent"
    elif ctx.get("test", False):
        result["comment"] = f"Would detach aws.organizations.policy_attachment {name}"
        return result
    else:
        try:
            result["old_state"] = before_ret["ret"]
            ret = await hub.exec.boto3.client.organizations.detach_policy(
                ctx,
                PolicyId=policy_id,
                TargetId=target_id,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = f"Detached '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes AWS Organization Policy attachments in a way that can be recreated/managed with the corresponding "present" function.

    This operation can be called only from the organization's management account
    or by a member account that is a delegated administrator for an AWS service.

    The only supported policy types are ``SERVICE_CONTROL_POLICY`` and ``TAG_POLICY``.

    Returns:
        Dict[str, Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.organizations.policy_attachment
    """
    result = {}

    ret = await hub.exec.boto3.client.organizations.list_policies(
        ctx, Filter="SERVICE_CONTROL_POLICY"
    )

    if not ret or not ret["ret"]:
        hub.log.debug(f"Could not describe policy_attachments {ret['comment']}")
        return result

    all_policies = ret["ret"].get("Policies", [])

    list_tag_policies = await hub.exec.boto3.client.organizations.list_policies(
        ctx, Filter="TAG_POLICY"
    )
    if not list_tag_policies:
        hub.log.debug(
            f"Could not describe policy_attachments {list_tag_policies['comment']}"
        )
        return {}

    if list_tag_policies["ret"]["Policies"]:
        all_policies.extend(list_tag_policies["ret"]["Policies"])

    if all_policies:

        for policy in all_policies:

            policy_attachments_ret = (
                await hub.exec.boto3.client.organizations.list_targets_for_policy(
                    ctx, PolicyId=policy["Id"]
                )
            )

            if not policy_attachments_ret["result"]:
                hub.log.warning(
                    f"Could not get attached target list with policy {policy['Id']} with error"
                    f" {policy_attachments_ret['comment']} . Describe will skip this policy and continue."
                )
            else:
                targets = policy_attachments_ret["ret"].get("Targets", [])

                if not targets:
                    hub.log.warning(
                        f"Attached target list with policy {policy['Id']} is empty."
                        f"Describe will skip this policy and continue."
                    )
                    continue

                for target in targets:
                    resource_name = f"{target['TargetId']}-{policy['Id']}"
                    translated_target = list()
                    translated_target.append({"policy_id": policy["Id"]})
                    translated_target.append({"target_id": target["TargetId"]})

                    result[resource_name] = {
                        "aws.organizations.policy_attachment.present": translated_target
                    }

    return result



import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    log_group_name: str,
    filter_pattern: str,
    destination_arn: str,
    resource_id: str = None,
    role_arn: str = None,
    distribution: str = None,
) -> Dict[str, Any]:
    r"""

    Creates or updates a subscription filter and associates it with the specified log group. Subscription filters
    allow you to subscribe to a real-time stream of log events ingested through PutLogEvents and have them delivered
    to a specific destination. When log events are sent to the receiving service, they are Base64 encoded and
    compressed with the gzip format. The following destinations are supported for subscription filters:   An Amazon
    Kinesis stream belonging to the same account as the subscription filter, for same-account delivery.   A logical
    destination that belongs to a different account, for cross-account delivery.   An Amazon Kinesis Firehose
    delivery stream that belongs to the same account as the subscription filter, for same-account delivery.   An
    Lambda function that belongs to the same account as the subscription filter, for same-account delivery.   Each
    log group can have up to two subscription filters associated with it. If you are updating an existing filter,
    you must specify the correct name in filterName.  To perform a PutSubscriptionFilter operation, you must also
    have the iam:PassRole permission.

    Args:
        name(Text): An Idem name of the resource.
        log_group_name(Text): The name of the log group.
        filter_pattern(Text): A filter pattern for subscribing to a filtered stream of log events.
        destination_arn(Text): The ARN of the destination to deliver matching log events to. Currently, the supported
            destinations are:   An Amazon Kinesis stream belonging to the same account as the subscription
            filter, for same-account delivery.   A logical destination (specified using an ARN) belonging to
            a different account, for cross-account delivery. If you are setting up a cross-account
            subscription, the destination must have an IAM policy associated with it that allows the sender
            to send logs to the destination. For more information, see PutDestinationPolicy.   An Amazon
            Kinesis Firehose delivery stream belonging to the same account as the subscription filter, for
            same-account delivery.   A Lambda function belonging to the same account as the subscription
            filter, for same-account delivery.
        resource_id(Text, optional): AWS logs Subscription filter name. Defaults to None.
        role_arn(Text, optional): The ARN of an IAM role that grants CloudWatch Logs permissions to deliver ingested log events to
            the destination stream. You don't need to provide the ARN when you are working with a logical
            destination for cross-account delivery. Defaults to None.
        distribution(Text, optional): The method used to distribute log data to the destination. By default, log data is grouped by
            log stream, but the grouping can be set to random for a more even distribution. This property is
            only applicable when the destination is an Amazon Kinesis stream. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.logs.subscription_filter.present:
                - name: value
                - log_group_name: value
                - filter_name: value
                - filter_pattern: value
                - role_arn: value
                - destination_arn: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    current_state = {}
    desired_state = {
        "name": name,
        "resource_id": resource_id,
        "log_group_name": log_group_name,
        "filter_pattern": filter_pattern,
        "destination_arn": destination_arn,
        "role_arn": role_arn,
        "distribution": distribution,
    }
    resource_updated = False
    if resource_id:
        ret = await hub.exec.boto3.client.logs.describe_subscription_filters(
            ctx, logGroupName=log_group_name, filterNamePrefix=resource_id
        )
        if not ret["result"]:
            result["result"] = False
            result["comment"] = ret["comment"]
            return result
        if ret["ret"]["subscriptionFilters"]:
            before = ret["ret"]["subscriptionFilters"][0]

    if before:
        current_state = hub.tool.aws.logs.subscription_filter_utils.convert_raw_subscription_filter_to_present(
            before, name
        )
        result["old_state"] = current_state

        resource_updated = (
            hub.tool.aws.logs.subscription_filter_utils.is_subscription_filter_updated(
                current_state, desired_state
            )
        )
        if not resource_updated:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.logs.subscription_filter", name=name
            )
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result
    if (not before) or resource_updated:
        if not ctx.get("test", False):
            ret = await hub.exec.boto3.client.logs.put_subscription_filter(
                ctx,
                logGroupName=log_group_name,
                filterName=name,
                filterPattern=filter_pattern,
                destinationArn=destination_arn,
                roleArn=role_arn,
                distribution=distribution,
            )
            if not ret["result"]:
                result["result"] = ret["result"]
                result["comment"] = ret["comment"]
                return result
            if resource_updated:
                result["comment"] = hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.logs.subscription_filter", name=name
                )
            else:
                result["comment"] = hub.tool.aws.comment_utils.create_comment(
                    resource_type="aws.logs.subscription_filter", name=name
                )
        else:
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state=current_state,
                desired_state=desired_state,
            )
            if resource_updated:
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.logs.subscription_filter", name=name
                )
            else:
                result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.logs.subscription_filter", name=name
                )
            return result

    if (not before) or resource_updated:
        resource_id = resource_id if resource_id else name
        ret = await hub.exec.boto3.client.logs.describe_subscription_filters(
            ctx, logGroupName=log_group_name, filterNamePrefix=resource_id
        )
        if not ret["result"]:
            result["result"] = False
            result["comment"] = ret["comment"]
            return result

        if ret["ret"]["subscriptionFilters"]:
            after = ret["ret"]["subscriptionFilters"][0]
        result[
            "new_state"
        ] = hub.tool.aws.logs.subscription_filter_utils.convert_raw_subscription_filter_to_present(
            after, name
        )

    return result


async def absent(
    hub,
    ctx,
    name: str,
    log_group_name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""

    Deletes the specified subscription filter.

    Args:
        name(Text): An Idem name of the resource.
        log_group_name(Text): The name of the log group.
        resource_id(Text, optional): AWS logs Subscription filter name. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.logs.subscription_filter.absent:
                - name: value
                - resource_id: value
                - log_group_name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.logs.subscription_filter", name=name
        )
        return result
    ret = await hub.exec.boto3.client.logs.describe_subscription_filters(
        ctx, logGroupName=log_group_name, filterNamePrefix=resource_id
    )
    if not ret["result"]:
        result["result"] = False
        result["comment"] = ret["comment"]
        return result
    if ret["ret"]["subscriptionFilters"]:
        before = ret["ret"]["subscriptionFilters"][0]

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.logs.subscription_filter", name=name
        )
    else:
        result[
            "old_state"
        ] = hub.tool.aws.logs.subscription_filter_utils.convert_raw_subscription_filter_to_present(
            raw_resource=before, idem_resource_name=name
        )
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.logs.subscription_filter", name=name
            )
            return result
        ret = await hub.exec.boto3.client.logs.delete_subscription_filter(
            ctx, logGroupName=log_group_name, filterName=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.logs.subscription_filter", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the subscription filters for all the log groups.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.logs.subscription_filter
    """

    result = {}
    log_group_ret = await hub.exec.boto3.client.logs.describe_log_groups(ctx)
    if not log_group_ret["result"]:
        hub.log.debug(f"Could not describe log groups {log_group_ret['comment']}")
        return {}

    for log_group in log_group_ret["ret"]["logGroups"]:
        ret = await hub.exec.boto3.client.logs.describe_subscription_filters(
            ctx, logGroupName=log_group["logGroupName"]
        )
        if not ret["result"]:
            hub.log.debug(
                f"Could not describe subscription_filters in log group {log_group['logGroupName']}.Will skip this log group {ret['comment']}"
            )
            continue
        for subscription_filter in ret["ret"]["subscriptionFilters"]:
            resource_id = subscription_filter.get("filterName")
            translated_resource = hub.tool.aws.logs.subscription_filter_utils.convert_raw_subscription_filter_to_present(
                subscription_filter, resource_id
            )
            result[resource_id] = {
                "aws.logs.subscription_filter.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in translated_resource.items()
                ]
            }
    return result



from typing import Any
from typing import Dict


async def update(
    hub,
    ctx,
    resource_id: str,
    before: Dict[str, Any],
    group_name: str,
    path: str,
):
    """
    Updates the IAM Group name and Path.

    Args:
        hub: required for functions in hub.
        ctx: context.
        resource_id: AWS IAM Group name.
        before(Dict): Existing resource parameters in Amazon Web Services.
        group_name(Str): IAM Group name
        path(str): Path Name of IAM group

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}

    if before.get("group_name") != group_name:
        update_payload["NewGroupName"] = group_name
    if before.get("path") != path:
        update_payload["NewPath"] = path

    if update_payload:
        result["ret"] = {}
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.iam.update_group(
                ctx=ctx, GroupName=resource_id, **update_payload
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result

            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.iam.group", name=resource_id
            )
        if "NewGroupName" in update_payload:
            result["ret"]["group_name"] = update_payload["NewGroupName"]
        if "NewPath" in update_payload:
            result["ret"]["path"] = update_payload["NewPath"]

    return result



import copy
from typing import Any
from typing import Dict


async def update_tags(
    hub,
    ctx,
    resource_arn: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """
    Update tags of AWS Lambda resources

    Args:
        resource_arn: Identifies the Amazon Lambda Function resource to which tags should be added.
            This value is an Amazon Resource Name (ARN).
        old_tags: List of existing tags
        new_tags: List of new tags

    Returns:
        {"result": True|False, "comment": ("A message",) "ret": None}

    """

    result = dict(comment=(), result=True, ret=None)
    tags_to_add = {}
    tags_to_delete = []
    tags_result = copy.deepcopy(old_tags)
    for key, value in new_tags.items():
        if (key in old_tags and old_tags.get(key) != new_tags.get(key)) or (
            key not in old_tags
        ):
            tags_to_add[key] = value

    for key in old_tags:
        if key not in new_tags:
            tags_to_delete.append(key)

    if (not tags_to_delete) and (not tags_to_add):
        return result

    if tags_to_delete:
        if not ctx.get("test", False):
            delete_tag_resp = await hub.exec.boto3.client["lambda"].untag_resource(
                ctx, Resource=resource_arn, TagKeys=tags_to_delete
            )
            if not delete_tag_resp:
                failure_message = (
                    f"Could not delete tags {tags_to_delete} for Lambda Function with ARN {resource_arn}."
                    f" Failed with error: {delete_tag_resp['comment']}"
                )
                hub.log.debug(failure_message)
                result["comment"] = (failure_message,)
                result["result"] = False
                return result
        [tags_result.pop(key, None) for key in tags_to_delete]

    if tags_to_add:
        if not ctx.get("test", False):
            create_tag_resp = await hub.exec.boto3.client["lambda"].tag_resource(
                ctx, Resource=resource_arn, Tags=tags_to_add
            )
            if not create_tag_resp:
                failure_message = (
                    f"Could not create tags {tags_to_add} for Lambda Function with ARN {resource_arn}."
                    f" Failed with error: {create_tag_resp['comment']}"
                )
                hub.log.debug(failure_message)
                result["comment"] = (failure_message,)
                result["result"] = False
                return result

    result["ret"] = {"tags": list(tags_result.values()) + list(tags_to_add)}
    result["comment"] = result["comment"] + (
        f"Update tags: Add [{tags_to_add}] Remove [{tags_to_delete}]",
    )
    return result



"""State module for managing S3 bucket's logging parameters."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]

TREQ = {
    "present": {"require": ["aws.s3.bucket.present"]},
}


async def present(
    hub,
    ctx,
    name: str,
    bucket_logging_status: make_dataclass(
        "BucketLoggingStatus",
        [
            (
                "LoggingEnabled",
                make_dataclass(
                    "LoggingEnabled",
                    [
                        ("TargetBucket", str),
                        ("TargetPrefix", str),
                        (
                            "TargetGrants",
                            List[
                                make_dataclass(
                                    "TargetGrant",
                                    [
                                        (
                                            "Grantee",
                                            make_dataclass(
                                                "Grantee",
                                                [
                                                    ("Type", str),
                                                    (
                                                        "DisplayName",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "EmailAddress",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                    ("ID", str, field(default=None)),
                                                    ("URI", str, field(default=None)),
                                                ],
                                            ),
                                            field(default=None),
                                        ),
                                        ("Permission", str, field(default=None)),
                                    ],
                                )
                            ],
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            )
        ],
    ),
    resource_id: str = None,
) -> Dict[str, Any]:
    """Create/Update the logging parameter for S3 bucket.

    Set the logging parameters for a bucket and to specify permissions for who can view and modify the logging
    parameters. All logs are saved to bucket in the same Amazon Web Services Region as the source bucket. To set
    the logging status of a bucket, you must be the bucket owner. The bucket owner is automatically granted
    FULL_CONTROL to all logs. You use the Grantee request element to grant access to other people. The Permissions
    request element specifies the kind of access the grantee has to the logs.

    Args:
        name(str):
            An Idem name of the resource.
            This is also used as the name of the bucket for which to set the logging parameters during resource creation.

        resource_id(str, optional):
            Name of the S3 bucket.

        bucket_logging_status(dict):
            Container for logging status information.

            * LoggingEnabled (dict, optional):
                Describes where logs are stored and the prefix that Amazon S3 assigns to all log object keys for a bucket.

                * TargetBucket (str):
                    Specifies the bucket where you want Amazon S3 to store server access logs.
                    You can have your logs delivered to any bucket that you own, including the same bucket that is being
                    logged. You can also configure multiple buckets to deliver their logs to the same target bucket.
                    In this case, you should choose a different TargetPrefix for each source bucket so that the delivered
                    log files can be distinguished by key.

                * TargetGrants (list, optional):
                    Container for granting information.
                    Buckets that use the bucket owner enforced setting for Object Ownership don't support target grants.

                    * Grantee (dict, optional):
                        Container for the person being granted permissions.

                        * DisplayName (str, optional):
                            Screen name of the grantee.

                        * EmailAddress (str, optional):
                            Email address of the grantee.

                        * ID (str, optional):
                            The canonical user ID of the grantee.

                        * Type (str):
                            Type of grantee.

                        * URI (str, optional):
                            URI of the grantee group.

                    * Permission (str, optional):
                        Logging permissions assigned to the grantee for the bucket.

                * TargetPrefix (str):
                    A prefix for all log object keys. If you store log files from multiple Amazon S3 buckets in a
                    single bucket, you can use a prefix to distinguish which log files came from which bucket.

    Request Syntax:
        .. code-block:: sls

            [bucket_name]-logging:
              aws.s3.bucket_logging.present:
              - name: "string"
              - resource_id: "string"
              - bucket_logging_status:
                  LoggingEnabled:
                    TargetBucket: "string"
                    TargetGrants:
                    - Grantee:
                        DisplayName: [string]
                        EmailAddress: [string]
                        ID: [string]
                        Type: [string]
                        URI: [string]
                      Permission: [string]
                    TargetPrefix: [string]

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            test-bucket-logging:
              aws.s3.bucket_logging.present:
                - name: test-bucket
                - bucket_logging_status:
                    LoggingEnabled:
                        TargetBucket: test-bucket
                        TargetGrants:
                        - Grantee:
                            URI: http://acs.amazonaws.com/groups/s3/LogDelivery
                            Type: Group
                          Permission: WRITE
                        - Grantee:
                            URI: http://acs.amazonaws.com/groups/s3/LogDelivery
                            Type: Group
                          Permission: READ_ACP
                        TargetPrefix: logs
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.s3.bucket_logging.get(
            ctx, resource_id=resource_id, name=name
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.s3.bucket_logging", name=name
        )
        result["old_state"] = copy.deepcopy(before["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])

        update_ret = await hub.tool.aws.s3.bucket_logging.update(
            ctx,
            resource_id=resource_id,
            before=before["ret"],
            bucket_logging_status=bucket_logging_status,
        )
        result["comment"] += update_ret["comment"]
        if not update_ret["result"]:
            result["result"] = False
            return result

        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            result["new_state"].update(update_ret["ret"])
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.s3.bucket_logging", name=name
            )
        if resource_updated and ctx.get("test", False):
            return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": name,
                    "bucket_logging_status": bucket_logging_status,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.s3.bucket_logging", name=name
            )
            return result
        create_ret = await hub.exec.boto3.client.s3.put_bucket_logging(
            ctx, Bucket=name, BucketLoggingStatus=bucket_logging_status
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.s3.bucket_logging", name=name
        )
        resource_id = name

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.aws.s3.bucket_logging.get(
            ctx, name=name, resource_id=resource_id
        )
        if not after_ret["result"] or not after_ret["ret"]:
            result["result"] = False
            result["comment"] += tuple(after_ret["comment"])
            return result
        result["new_state"] = after_ret["ret"]

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes the logging configuration for an S3 bucket.

    This action enables you to delete logging configurations from a bucket using a single HTTP request.
    If you know the object keys that you want to delete, then this action provides a suitable alternative to sending
    individual delete requests, reducing per-request overhead. For each configuration, Amazon S3 performs a delete
    action and returns the result of that delete, success, or failure, in the response.

    Args:
        name(str): Idem name of the bucket.
        resource_id(Text): Name of the bucket on which logging needs to be deleted.

    Request Syntax:
        .. code-block:: sls

            [bucket_name]-logging:
              aws.s3.bucket_logging.absent:
                - name: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            test-bucket-logging:
              aws.s3.bucket_logging.absent:
                - name: test-bucket
                - resource_id: test-bucket
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.s3.bucket_logging", name=name
        )
        return result

    before = await hub.exec.aws.s3.bucket_logging.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.s3.bucket_logging", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.s3.bucket_logging", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        delete_ret = await hub.exec.boto3.client.s3.put_bucket_logging(
            ctx, Bucket=resource_id, BucketLoggingStatus={}
        )
        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.s3.bucket_logging", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Gets information about the S3 bucket logging parameters.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.s3.bucket_logging

    """
    result = {}
    # To describe the logging configurations of all the buckets, we first need to list all the buckets,
    # then get the logging configurations of each bucket
    ret = await hub.exec.boto3.client.s3.list_buckets(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe S3 buckets {ret['comment']}")
        return {}

    for bucket in ret["ret"]["Buckets"]:
        bucket_name = bucket.get("Name")
        # get logging configuration for each bucket
        bucket_logging_response = await hub.exec.boto3.client.s3.get_bucket_logging(
            ctx, Bucket=bucket_name
        )
        if bucket_logging_response["result"]:
            if (
                bucket_logging_response["ret"]
                and "LoggingEnabled" in bucket_logging_response["ret"]
            ):
                bucket_logging_response["ret"].pop("ResponseMetadata", None)
                translated_resource = hub.tool.aws.s3.conversion_utils.convert_raw_bucket_logging_to_present(
                    ctx=ctx,
                    raw_resource=bucket_logging_response["ret"],
                    bucket_name=bucket_name,
                )

                result[bucket_name + "-logging"] = {
                    "aws.s3.bucket_logging.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in translated_resource.items()
                    ]
                }
    return result



from typing import Dict


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    load_balancer_arn: str = None,
) -> Dict:
    """
    Pass required params to get a target group resource.

    **** Note ****: Users can specify one of the following to filter the results: the ARN of the load balancer,
                    name of the target group, or the ARN of the given target group.

    Order of precedence of input params while performing search: resource_id, load_balancer_arn and finally name.
    If resource_id is not None, regardless whether the 2 other values are None or not, search always is done by
    resource_id only.

    Args:
        name(Text): The name of the AWS ELBv2 Target Group. This name must be unique per region per account.
        resource_id(Text, optional): AWS ELBv2 Target Group ARN to identify the resource.
        load_balancer_arn(Text, optional): The Amazon Resource Name (ARN) of the load balancer.

    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.elbv2.target_group.search_raw(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        load_balancer_arn=load_balancer_arn,
    )
    if not ret["result"]:
        if "TargetGroupNotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.elbv2.target_group", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] = list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["TargetGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.elbv2.target_group", name=name
            )
        )
        return result

    resource_id = ret["ret"]["TargetGroups"][0].get("TargetGroupArn")
    tags = attributes = targets = []
    if resource_id:
        tags_ret = await hub.exec.boto3.client.elbv2.describe_tags(
            ctx, ResourceArns=[resource_id]
        )
        if not tags_ret["result"]:
            result["comment"] = list(tags_ret["comment"])
            result["result"] = False
            return result
        else:
            if tags_ret["ret"] and tags_ret.get("ret")["TagDescriptions"]:
                tags = (tags_ret["ret"]["TagDescriptions"][0]).get("Tags")

        attributes_ret = (
            await hub.exec.boto3.client.elbv2.describe_target_group_attributes(
                ctx, TargetGroupArn=resource_id
            )
        )
        if not attributes_ret.get("result"):
            result["comment"] = list(attributes_ret["comment"])
            result["result"] = False
            return result
        else:
            if attributes_ret.get("ret") and attributes_ret.get("ret").get(
                "Attributes"
            ):
                attributes = attributes_ret["ret"].get("Attributes")

        targets_ret = await hub.exec.boto3.client.elbv2.describe_target_health(
            ctx, TargetGroupArn=resource_id
        )
        if not targets_ret.get("result"):
            result["comment"] = list(targets_ret["comment"])
            result["result"] = False
            return result
        else:
            if targets_ret.get("ret") and targets_ret["ret"].get(
                "TargetHealthDescriptions"
            ):
                targets = targets_ret["ret"].get("TargetHealthDescriptions")

    result[
        "ret"
    ] = hub.tool.aws.elbv2.conversion_utils.convert_raw_target_group_to_present(
        raw_resource=ret["ret"]["TargetGroups"][0],
        idem_resource_name=name,
        tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags),
        attributes=attributes,
        targets=targets,
    )
    return result



import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.s3.bucket.present",
        ],
    },
    "absent": {
        "require": [
            "aws.s3.bucket.absent",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    bucket: str,
    public_access_block_configuration: make_dataclass(
        "PublicAccessBlockConfiguration",
        [
            ("BlockPublicAcls", bool, field(default=None)),
            ("IgnorePublicAcls", bool, field(default=None)),
            ("BlockPublicPolicy", bool, field(default=None)),
            ("RestrictPublicBuckets", bool, field(default=None)),
        ],
    ),
    expected_bucket_owner: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""

    Creates or modifies the PublicAccessBlock configuration for an Amazon S3 bucket.
    To use this operation, the calling entity must have the s3:PutBucketPublicAccessBlock permission.

    Args:
      hub:
      ctx:
      name(Text): The name of the bucket public access block configuration.
      bucket(Text): Bucket name to identify the resource
        public_access_block_configuration(Dict[str, Any]): The PublicAccessBlock configuration that you want to apply to this Amazon S3 bucket. You can
            enable the configuration options in any combination. For more information about when Amazon S3
            considers a bucket or object public, see The Meaning of "Public" in the Amazon S3 User Guide.
            * BlockPublicAcls (bool, optional): Specifies whether Amazon S3 should block public access control lists (ACLs) for this bucket and
                objects in this bucket. Setting this element to TRUE causes the following behavior:   PUT Bucket
                ACL and PUT Object ACL calls fail if the specified ACL is public.   PUT Object calls fail if the
                request includes a public ACL.   PUT Bucket calls fail if the request includes a public ACL.
                Enabling this setting doesn't affect existing policies or ACLs.
            * IgnorePublicAcls (bool, optional): Specifies whether Amazon S3 should ignore public ACLs for this bucket and objects in this
                bucket. Setting this element to TRUE causes Amazon S3 to ignore all public ACLs on this bucket
                and objects in this bucket. Enabling this setting doesn't affect the persistence of any existing
                ACLs and doesn't prevent new public ACLs from being set.
            * BlockPublicPolicy (bool, optional): Specifies whether Amazon S3 should block public bucket policies for this bucket. Setting this
                element to TRUE causes Amazon S3 to reject calls to PUT Bucket policy if the specified bucket
                policy allows public access.  Enabling this setting doesn't affect existing bucket policies.
            * RestrictPublicBuckets (bool, optional): Specifies whether Amazon S3 should restrict public bucket policies for this bucket. Setting this
                element to TRUE restricts access to this bucket to only Amazon Web Service principals and
                authorized users within this account if the bucket has a public policy. Enabling this setting
                doesn't affect previously stored bucket policies, except that public and cross-account access
                within any public bucket policy, including non-public delegation to specific accounts, is
                blocked.
      expected_bucket_owner (Text, optional): The account ID of the expected bucket owner. If the bucket is owned by a different account, the request will fail with an HTTP 403 (Access Denied) error.
      resource_id (Text, Optional): Bucket name to identify the resource

    Request Syntax:
      resource_name:
        aws.s3.public_access_block.present:
          - name: string
          - bucket: string
          - public_access_block_configuration: Dict
          - expected_bucket_owner: string
          - resource_id: string

    Returns:
      Dict[str, Any]

    Examples:

      .. code-block:: sls

      test-bucket-1232323-public-access-block:
        aws.s3.public_access_block.present:
          - name: test-bucket-1232323-public-access-block
          - bucket: test-bucket-1232323
          - public_access_block_configuration: {"BlockPublicAcls": true, "IgnorePublicAcls": true, "BlockPublicPolicy": true, "RestrictPublicBuckets": true }
          - expected_bucket_owner: 1234567890
          - resource_id: test-bucket-1232323


    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_id = resource_id if resource_id else bucket
    # fetching old public access block configuration
    old_configuration_ret = await hub.exec.boto3.client.s3.get_public_access_block(
        ctx, Bucket=resource_id
    )
    if old_configuration_ret["result"]:
        old_public_access_block_configuration = old_configuration_ret["ret"][
            "PublicAccessBlockConfiguration"
        ]
        result[
            "old_state"
        ] = hub.tool.aws.s3.conversion_utils.convert_raw_public_access_block_to_present(
            bucket, old_public_access_block_configuration
        )
    elif "NoSuchPublicAccessBlockConfiguration" in str(
        old_configuration_ret["comment"]
    ):
        old_public_access_block_configuration = None
    else:
        result["comment"] = old_configuration_ret["comment"]
        result["result"] = False
        return result

    # If new public access block configuration is same as old
    if old_public_access_block_configuration == public_access_block_configuration:
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] = (f"aws.s3.public_access_block '{name}' already exists",)
        return result

    if ctx.get("test", False):
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state={},
            desired_state={
                "name": name,
                "bucket": bucket,
                "public_access_block_configuration": public_access_block_configuration,
                "expected_bucket_owner": expected_bucket_owner,
            },
        )
        operation = "create" if not old_public_access_block_configuration else "update"
        result[
            "comment"
        ] = f"Would {operation} aws.s3.public_access_block '{name}' configuration for bucket {bucket}"
        return result

    # create or update the bucket public access block configuration
    ret = await hub.exec.boto3.client.s3.put_public_access_block(
        ctx,
        Bucket=resource_id,
        PublicAccessBlockConfiguration=public_access_block_configuration,
        ExpectedBucketOwner=expected_bucket_owner,
    )

    if not ret["result"]:
        result["comment"] = result["comment"] + ret["comment"]
        result["result"] = False
        return result

    # fetching new public access block configuration
    new_configuration_ret = await hub.exec.boto3.client.s3.get_public_access_block(
        ctx, Bucket=resource_id
    )
    result[
        "new_state"
    ] = hub.tool.aws.s3.conversion_utils.convert_raw_public_access_block_to_present(
        bucket, new_configuration_ret["ret"]["PublicAccessBlockConfiguration"]
    )
    operation = "Created" if not old_public_access_block_configuration else "Updated"
    result["comment"] = (f"{operation} aws.s3.public_access_block '{name}'",)
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
    expected_bucket_owner: str = None,
) -> Dict[str, Any]:
    r"""

    Removes the PublicAccessBlock configuration for an Amazon S3 bucket.
    To use this operation, you must have the s3:PutBucketPublicAccessBlock permission.

    Args:
        name(Text): The name of public block configuration
        resource_id (Text): Bucket name to identify the resource
        expected_bucket_owner (Text, optional): The account ID of the expected bucket owner. If the bucket is owned by a different account, the request will fail with an HTTP 403 (Access Denied) error.

    Returns:
        Dict[str, Any]

    Request Syntax:
        resource_name:
          aws.s3.public_access_block.absent:
            - name: string
            - resource_id: string
            - expected_bucket_owner: string

    Examples:

        .. code-block:: sls

            test-bucket-1232323-public-access-block:
              aws.s3.public_access_block.absent:
              - name: test-bucket-1232323-public-access-block
              - resource_id: test-bucket-1232323
              - expected_bucket_owner: 1234567890
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # fetching old public access block configuration
    old_configuration_ret = await hub.exec.boto3.client.s3.get_public_access_block(
        ctx, Bucket=resource_id
    )

    if old_configuration_ret["result"]:
        old_public_access_block_configuration = old_configuration_ret["ret"][
            "PublicAccessBlockConfiguration"
        ]
        result[
            "old_state"
        ] = hub.tool.aws.s3.conversion_utils.convert_raw_public_access_block_to_present(
            resource_id, old_public_access_block_configuration
        )
    elif "NoSuchPublicAccessBlockConfiguration" in str(
        old_configuration_ret["comment"]
    ):
        result["comment"] = (f"aws.s3.public_access_block '{name}' already absent",)
        return result
    else:
        result["comment"] = old_configuration_ret["comment"]
        result["result"] = False
        return result

    if ctx.get("test", False):
        result[
            "comment"
        ] = f"Would delete aws.s3.public_access_block '{name}' configuration for bucket {resource_id}"
        return result

    ret = await hub.exec.boto3.client.s3.delete_public_access_block(
        ctx, Bucket=resource_id, ExpectedBucketOwner=expected_bucket_owner
    )
    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"] = ret["comment"]
        result["result"] = False
        return result
    result[
        "comment"
    ] = f"Deleted aws.s3.public_access_block '{name}' for bucket '{resource_id}'"

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Obtain S3 public access block configuration for each bucket under the given context for any user.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.s3.public_access_block
    """
    result = {}
    # To describe the public access block configuration of all the buckets, we first need to list all the buckets, then get the
    # configuration of each bucket
    ret_buckets = await hub.exec.boto3.client.s3.list_buckets(ctx)

    if not ret_buckets["result"]:
        hub.log.debug(f"Could not describe buckets {ret_buckets['comment']}")
        return result

    for bucket in ret_buckets["ret"]["Buckets"]:
        bucket_name = bucket.get("Name")
        resource_name = f"{bucket_name}-public-access-block"
        # get configuration for each bucket
        config_ret = await hub.exec.boto3.client.s3.get_public_access_block(
            ctx, Bucket=bucket_name
        )
        if not config_ret["result"]:
            hub.log.warning(
                f"Could not get public access block configuration for bucket {bucket_name} with error"
                f" {config_ret['comment']} . Describe will skip this bucket and continue."
            )
        else:
            resource_translated = hub.tool.aws.s3.conversion_utils.convert_raw_public_access_block_to_present(
                bucket_name, config_ret["ret"].get("PublicAccessBlockConfiguration")
            )
            result[resource_name] = {
                "aws.s3.public_access_block.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


async def update(
    hub,
    ctx,
    before: Dict[str, Any],
    account_id: str,
    notification_type: str,
    action_threshold: Dict[str, Any],
    definition: Dict[Any, Any],
    execution_role_arn: str,
    approval_model: str,
    subscribers: List[Dict[str, Any]],
):
    """
    Updates the Budget Action
    Args:
        before: existing resource
        account_id(Text): Account ID
        notification_type(Text): The type of a notification. It must be ACTUAL or FORECASTED.
        action_threshold(Dict): The trigger threshold of the action.
        definition(Dict[Any, Any]): Specifies all of the type-specific parameters. IamActionDefinition, ScpActionDefinition, SsmActionDefinition
        execution_role_arn(Text): The role passed for action execution and reversion. Roles and actions must be in the same account.
        approval_model (Text): This specifies if the action needs manual or automatic approval.
        subscribers (List[Dict[str, Any]]): A list of subscribers.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": None}

    """
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}

    resource_parameters = OrderedDict(
        {
            "NotificationType": notification_type,
            "ActionThreshold": action_threshold,
            "Definition": definition,
            "ExecutionRoleArn": execution_role_arn,
            "ApprovalModel": approval_model,
        }
    )

    for key, value in resource_parameters.items():
        if key in before.keys():
            if (value is not None) and value != before[key]:
                update_payload[key] = resource_parameters[key]

    if not hub.tool.aws.state_comparison_utils.are_lists_identical(
        subscribers,
        before.get("Subscribers"),
    ):
        update_payload["Subscribers"] = subscribers

    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.budgets.update_budget_action(
                ctx,
                AccountId=account_id,
                BudgetName=before.get("BudgetName"),
                ActionId=before.get("ActionId"),
                **update_payload,
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        result = update_result(result, update_payload)

    return result


def update_result(
    result: Dict[str, Any], update_payload: Dict[str, Any]
) -> Dict[str, Any]:
    parameters = OrderedDict(
        {
            "NotificationType": "notification_type",
            "ActionThreshold": "action_threshold",
            "Definition": "definition",
            "ExecutionRoleArn": "execution_role_arn",
            "ApprovalModel": "approval_model",
            "Subscribers": "subscribers",
        }
    )

    for raw_parameter, present_parameter in parameters.items():
        if raw_parameter in update_payload:
            result["ret"][present_parameter] = update_payload[raw_parameter]
            result["comment"] = result["comment"] + (
                f"Update {present_parameter}: {update_payload[raw_parameter]}",
            )
    return result



"""State module for managing Amazon ECR Repositories."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    registry_id: str = None,
    image_tag_mutability: str = None,
    image_scanning_configuration: make_dataclass(
        """The image scanning configuration for the repository."""
        "ImageScanningConfiguration",
        [("scanOnPush", bool, field(default=None))],
    ) = None,
    encryption_configuration: make_dataclass(
        """The encryption configuration for the repository."""
        "EncryptionConfiguration",
        [("encryptionType", str), ("kmsKey", str, field(default=None))],
    ) = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    """Creates an AWS ECR repository.

    Repository properties cannot be updated after creation, only tags can be updated.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            The name of the ECR repository in Amazon Web Services.
        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry to create the repository. If you
            do not specify a registry, the default registry is assumed.
        image_tag_mutability(str, Optional):
            The tag mutability setting for the repository. If this parameter is omitted, the default setting
            of ``MUTABLE`` will be used which will allow image tags to be overwritten. If ``IMMUTABLE`` is
            specified, all image tags within the repository will be immutable which will prevent them from
            being overwritten.
        image_scanning_configuration(dict, Optional):
            The image scanning configuration for the repository.

            * scanOnPush (*bool, Optional*):
                The setting that determines whether images are scanned after being pushed to a repository.
                If set to true , images will be scanned after being pushed. If this parameter is not specified,
                it will default to false and images will not be scanned unless a scan is manually started.
        encryption_configuration(dict, Optional):
            The encryption configuration for the repository.

            * encryptionType (*str*):
                The encryption type to use. If you use the ``KMS`` encryption type, the contents of the repository will be encrypted
                using server-side encryption with Key Management Service key stored in KMS. If you use the ``AES256`` encryption type,
                Amazon ECR uses server-side encryption with Amazon S3-managed encryption keys which encrypts the images in the
                repository using an AES-256 encryption algorithm.
            * kmsKey (*str, Optional*):
                If you use the KMS encryption type, specify the KMS key to use for encryption. The alias, key ID,
                or full ARN of the KMS key can be specified. The key must exist in the same Region as the repository.
                If no key is specified, the default Amazon Web Services managed KMS key for Amazon ECR will be used.
        tags(dict or list, Optional):
            Dict in the format of ``{tag-key: tag-value}`` or List of tags in the format of
            ``[{"Key": tag-key, "Value": tag-value}]`` to associate with the repository.

            * Key (*str*):
                One part of a key-value pair that make up a tag. A key is a general label that acts like a category for more specific
                tag values. Tag keys can have a maximum character length of 128 characters
            * Value (*str*):
                A value acts as a descriptor within a tag category (key). Tag values can have a maximum length of 256 characters.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_ecr_repository]:
          aws.ecr.repository.present:
            - name: 'string'
            - registry_id: 'string'
            - image_tag_mutability: 'MUTABLE|IMMUTABLE'
            - image_scanning_configuration:
                scanOnPush: True|False
            - encryption_configuration:
                encryptionType: 'AES256|KMS'
                kmsKey: 'string'
            - tags:
              - Key: 'string'
                Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_ecr_repository:
              aws.ecr.repository.present:
                - name: idem_test_ecr_repository
                - registry_id: idem_test_ecr_registry
                - image_tag_mutability: 'MUTABLE'
                - image_scanning_configuration:
                    scanOnPush: False
                - encryption_configuration:
                    encryptionType: 'AES256'
                - tags:
                  - Key: provider
                    Value: idem
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = None
    if resource_id:
        before_ret = await hub.exec.boto3.client.ecr.describe_repositories(
            ctx, repositoryNames=[resource_id], registryId=registry_id
        )
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    # Repositories cannot be updated after creation with the Amazon Web Service API,
    # only repository tags can be updated
    if before_ret and before_ret["ret"].get("repositories"):  # resource present
        result["comment"] = (f"aws.ecr.repository '{name}' already exists",)

        updated = False
        repository = before_ret["ret"].get("repositories")[0]

        before_tags_ret = await hub.exec.boto3.client.ecr.list_tags_for_resource(
            ctx, resourceArn=repository.get("repositoryArn")
        )
        if not before_tags_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + before_tags_ret["comment"]
            return result

        result[
            "old_state"
        ] = hub.tool.aws.ecr.conversion_utils.convert_raw_repository_to_present(
            raw_resource=repository,
            tags=before_tags_ret.get("ret").get("tags")
            if before_tags_ret.get("result")
            else None,
            idem_resource_name=name,
        )

        if tags is not None and tags != result["old_state"].get("tags"):
            update_tags_ret = await hub.tool.aws.ecr.tag.update_tags(
                ctx=ctx,
                resource_arn=repository.get("repositoryArn"),
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )

            if not update_tags_ret["result"]:
                result["result"] = False
                result["comment"] = result["comment"] + update_tags_ret["comment"]
                return result

            updated = updated or bool(update_tags_ret["result"])
            result["comment"] = result["comment"] + update_tags_ret["comment"]

            if ctx.get("test", False) and updated:
                result["new_state"] = copy.deepcopy(result["old_state"])
                if len(update_tags_ret["ret"]) > 0:
                    result["new_state"]["tags"] = update_tags_ret["ret"]
                else:
                    result["new_state"]["tags"] = None
                result["comment"] = result["comment"] + (
                    f"Would have updated tags for aws.ecr.repository '{name}'",
                )
                return result

        if not updated:
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result

        result["comment"] = result["comment"] + (
            f"Updated aws.ecr.repository '{name}'",
        )
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": name,
                    "registry_id": registry_id,
                    "repository_name": name,
                    "image_tag_mutability": image_tag_mutability,
                    "image_scanning_configuration": image_scanning_configuration,
                    "encryption_configuration": encryption_configuration,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ecr.repository", name=name
            )
            return result

        create_ret = await hub.exec.boto3.client.ecr.create_repository(
            ctx,
            registryId=registry_id,
            repositoryName=name,
            imageTagMutability=image_tag_mutability,
            imageScanningConfiguration=image_scanning_configuration,
            encryptionConfiguration=encryption_configuration,
            tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            if tags
            else None,
        )
        if not result["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ecr.repository", name=name
        )
        resource_id = name

    after_ret = await hub.exec.boto3.client.ecr.describe_repositories(
        ctx, registryId=registry_id, repositoryNames=[resource_id]
    )
    if not after_ret["result"]:
        result["result"] = False
        result["comment"] = after_ret["comment"]
        return result

    repository = after_ret["ret"].get("repositories")[0]

    after_tags_ret = await hub.exec.boto3.client.ecr.list_tags_for_resource(
        ctx, resourceArn=repository.get("repositoryArn")
    )
    if not after_tags_ret["result"]:
        result["result"] = False
        result["comment"] = after_tags_ret["comment"]
        return result

    result[
        "new_state"
    ] = hub.tool.aws.ecr.conversion_utils.convert_raw_repository_to_present(
        raw_resource=repository,
        tags=after_tags_ret.get("ret").get("tags")
        if after_tags_ret.get("result")
        else None,
        idem_resource_name=name,
    )

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    registry_id: str = None,
    force: bool = None,
) -> Dict[str, Any]:
    """Deletes an AWS ECR repository.

    If the repository contains images, you must either delete all images in the repository or
    use the force option to delete the repository.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            The name of the ECR repository in Amazon Web Services.

            .. warning::
              Idem automatically considers this resource being absent if this field is not specified.

        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that contains the repository to
            delete. If you do not specify a registry, the default registry is assumed.
        force(bool, Optional):
            If a repository contains images, forces the deletion.

    Request syntax:
      .. code-block:: sls

        [idem_test_aws_ecr_repository]:
          aws.ecr.repository.absent:
            - name: 'string'
            - resource_id: 'string'
            - registry_id: 'string'
            - force: True|False

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_ecr_repository:
              aws.ecr.repository.absent:
                - name: idem_test_ecr_repository
                - resource_id: idem_test_ecr_repository
                - registry_id: idem_test_ecr_registry
                - force: True
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ecr.repository", name=name
        )
        return result

    before_ret = await hub.exec.boto3.client.ecr.describe_repositories(
        ctx, repositoryNames=[resource_id], registryId=registry_id
    )
    if not before_ret["result"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ecr.repository", name=name
        )
        return result

    repository = before_ret["ret"].get("repositories")[0]

    tags_ret = await hub.exec.boto3.client.ecr.list_tags_for_resource(
        ctx, resourceArn=repository.get("repositoryArn")
    )
    if not tags_ret["result"]:
        result["result"] = False
        result["comment"] = tags_ret["comment"]
        return result

    result[
        "old_state"
    ] = hub.tool.aws.ecr.conversion_utils.convert_raw_repository_to_present(
        raw_resource=repository,
        tags=tags_ret.get("ret").get("tags") if tags_ret.get("result") else None,
        idem_resource_name=name,
    )

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ecr.repository", name=name
        )
        return result

    delete_ret = await hub.exec.boto3.client.ecr.delete_repository(
        ctx, repositoryName=resource_id, registryId=registry_id, force=force
    )
    if not result["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.ecr.repository", name=name
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes AWS ECR repositories in a registry in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ecr.repository
    """
    result = {}

    describe_ret = await hub.exec.boto3.client.ecr.describe_repositories(ctx)
    if not describe_ret["result"]:
        hub.log.debug(f"Could not describe repository: {describe_ret['comment']}")
        return result

    for repository in describe_ret["ret"].get("repositories"):
        resource_id = repository.get("repositoryName")

        tags_ret = await hub.exec.boto3.client.ecr.list_tags_for_resource(
            ctx, resourceArn=repository.get("repositoryArn")
        )
        if not tags_ret["result"]:
            hub.log.debug(
                f"Could not list tags for repository '{resource_id}': {tags_ret['comment']}"
            )

        resource_translated = (
            hub.tool.aws.ecr.conversion_utils.convert_raw_repository_to_present(
                raw_resource=repository,
                tags=tags_ret.get("ret").get("tags")
                if tags_ret.get("result")
                else None,
                idem_resource_name=resource_id,
            )
        )

        result[resource_id] = {
            "aws.ecr.repository.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result


async def search(
    hub, ctx, name, resource_id: str, registry_id: str = None, default: bool = None
) -> Dict[str, Any]:
    """Use an un-managed AWS ECR repository as a data-source.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str):
            The name of the ECR repository in Amazon Web Services.
        registry_id(str, Optional):
            The Amazon Web Services account ID associated with the registry that contains the repository
            to search. If you do not specify a registry, the default registry is assumed.

    Request syntax:
      .. code-block:: sls

        [idem_test_aws_ecr_repository]:
          aws.ecr.repository.search:
            - name: 'string'
            - resource_id: 'string'
            - registry_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_ecr_repository:
              aws.ecr.repository.search:
                - name: idem_test_ecr_repository
                - resource_id: idem_test_ecr_repository
                - registry_id: idem_test_ecr_registry
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    describe_ret = await hub.exec.boto3.client.ecr.describe_repositories(
        ctx, repositoryNames=[resource_id], registryId=registry_id
    )
    if not describe_ret["result"]:
        result["result"] = False
        result["comment"] = (
            f"Unable to find aws.ecr.repository resource with repository name '{resource_id}'",
        )
        return result

    if len(describe_ret["ret"].get("repositories")) > 1:
        result["result"] = False
        result["comment"] = (
            f"More than one aws.ecr.repository resource with name '{resource_id}' was found",
        )
        return result

    repository = describe_ret["ret"].get("repositories")[0]

    tags_ret = await hub.exec.boto3.client.ecr.list_tags_for_resource(
        ctx, resourceArn=repository.get("repositoryArn")
    )
    if not tags_ret["result"]:
        result["result"] = False
        result["comment"] = tags_ret["comment"]
        return result

    result[
        "old_state"
    ] = hub.tool.aws.ecr.conversion_utils.convert_raw_repository_to_present(
        raw_resource=repository,
        tags=tags_ret.get("ret").get("tags") if tags_ret.get("result") else None,
        idem_resource_name=name,
    )
    result["new_state"] = copy.deepcopy(result["old_state"])

    return result
